from amoeba.Util.util import (
    create_maps,
    calculate_keplerian_velocity,
    convert_spin_to_isco_radius,
    convert_eddington_ratio_to_accreted_mass,
    accretion_disk_temperature,
    calculate_gravitational_radius,
    planck_law,
    planck_law_derivative,
    calculate_angular_diameter_distance,
    calculate_angular_diameter_distance_difference,
    calculate_luminosity_distance,
    calculate_angular_einstein_radius,
    calculate_einstein_radius_in_meters,
    pull_value_from_grid,
    convert_1d_array_to_2d_array,
    convert_cartesian_to_polar,
    convert_polar_to_cartesian,
    perform_microlensing_convolution,
    extract_light_curve,
    calculate_time_lag_array,
    calculate_geometric_disk_factor,
    calculate_dt_dlx,
    construct_accretion_disk_transfer_function,
    calculate_microlensed_transfer_function,
    generate_drw_signal,
    generate_signal_from_psd,
    generate_snapshots_of_radiation_pattern,
    project_blr_to_source_plane,
    calculate_blr_transfer_function,
    determine_emission_line_velocities,
    convolve_signal_with_transfer_function,
    convert_speclite_filter_to_wavelength_range,
)
import pytest
import numpy as np
import numpy.testing as npt
from astropy import units as u
from astropy import constants as const
from astropy.cosmology import FlatLambdaCDM
from speclite.filters import FilterResponse, load_filters


def test_create_maps():
    mass_exp = 7.0
    redshift = 1.0
    number_grav_radii = 100
    inc_ang = 45
    resolution = 100

    neg_redshift = -1.0
    bad_spin = 1.5
    bad_beta = -42
    bad_inc_ang = 175
    with pytest.raises(AssertionError):
        create_maps(mass_exp, neg_redshift, number_grav_radii, inc_ang, resolution)
    with pytest.raises(AssertionError):
        create_maps(
            mass_exp, redshift, number_grav_radii, inc_ang, resolution, spin=bad_spin
        )
    with pytest.raises(AssertionError):
        create_maps(mass_exp, redshift, number_grav_radii, bad_inc_ang, resolution)
    with pytest.raises(AssertionError):
        create_maps(
            mass_exp,
            neg_redshift,
            number_grav_radii,
            inc_ang,
            resolution,
            temp_beta=bad_beta,
        )
    map1 = create_maps(mass_exp, redshift, number_grav_radii, inc_ang, resolution)
    map2 = create_maps(
        mass_exp, redshift, number_grav_radii, inc_ang, 2 * resolution, albedo=0.4
    )
    # There should be 10 objects
    assert type(map1) == dict
    # The 4 array objects should be of resolution of the image
    assert np.size(map2["radii_array"]) == np.size(map1["radii_array"]) * 4

    edge_on_inc = 90
    albedo_array = np.zeros((resolution, resolution))
    map_testing = create_maps(
        mass_exp,
        redshift,
        number_grav_radii,
        edge_on_inc,
        resolution,
        albedo=albedo_array,
    )


def test_calculate_keplerian_velocity():
    radius_1 = 100
    radius_2 = 10000
    mass_1 = 10
    mass_2 = 20

    kep_vel_1_1 = calculate_keplerian_velocity(radius_1, mass_1)
    kep_vel_1_2 = calculate_keplerian_velocity(radius_1, mass_2)
    kep_vel_2_1 = calculate_keplerian_velocity(radius_2, mass_1)

    # check scaling with params
    assert kep_vel_1_1 == 10 * kep_vel_2_1
    assert kep_vel_1_1 == kep_vel_1_2 / np.sqrt(2)

    # check astropy units work
    mass_astropy_1 = 10 * const.M_sun.to(u.kg)
    radius_astropy_1 = 100 * u.m
    kep_vel_astropy = calculate_keplerian_velocity(radius_1, mass_astropy_1)
    assert kep_vel_1_1 == kep_vel_astropy

    # check that we are returning a numerical value (not an astropy unit).
    assert isinstance(kep_vel_1_1, float)


def test_convert_spin_to_isco_radius():
    spin_positive = 1
    spin_zero = 0
    spin_negative = -1

    # test known limits
    assert convert_spin_to_isco_radius(spin_positive) == 1
    assert convert_spin_to_isco_radius(spin_zero) == 6
    assert convert_spin_to_isco_radius(spin_negative) == 9

    # test out of bounds
    with pytest.raises(ValueError):
        convert_spin_to_isco_radius(4)
    with pytest.raises(ValueError):
        convert_spin_to_isco_radius(-20)


def test_convert_eddington_ratio_to_accreted_mass():
    mass = 10**7
    mass_quantity = 10**7 * const.M_sun.to(u.kg)
    eddington_ratio_1 = 0.1
    efficiency_1 = 0.1
    eddington_ratio_2 = 0.2
    efficiency_2 = 0.2

    accreted_mass_1_1 = convert_eddington_ratio_to_accreted_mass(
        mass, eddington_ratio_1, efficiency=efficiency_1
    )
    accreted_mass_1_2 = convert_eddington_ratio_to_accreted_mass(
        mass, eddington_ratio_1, efficiency=efficiency_2
    )
    accreted_mass_2_1 = convert_eddington_ratio_to_accreted_mass(
        mass, eddington_ratio_2, efficiency=efficiency_1
    )
    accreted_mass_quantity = convert_eddington_ratio_to_accreted_mass(
        mass_quantity, eddington_ratio_1, efficiency=efficiency_1
    )

    # test that this function has the proper scaling
    # e.g. twice as efficient means half the accreted mass
    # and twice the eddington ratio means twice the accreted mass
    assert (accreted_mass_1_1 / 2) == accreted_mass_1_2
    assert (accreted_mass_1_1 * 2) == accreted_mass_2_1
    # Check astropy quantity works as an input
    assert accreted_mass_1_1 == accreted_mass_quantity


def test_accretion_disk_temperature():
    mass_in_m_sun = 10**8
    grav_rad = calculate_gravitational_radius(mass_in_m_sun)
    radii_in_meters = np.linspace(10, 100, 90) * grav_rad
    min_radius_in_meters = 6 * grav_rad
    eddington_ratio = 0.1

    temp_profile_SS = accretion_disk_temperature(
        radii_in_meters,
        min_radius_in_meters,
        mass_in_m_sun,
        eddington_ratio,
        beta=0,
        corona_height=6,
        albedo=1,
        eta_x_rays=0.1,
        generic_beta=False,
        disk_acc=None,
        efficiency=0.1,
        spin=0,
        visc_temp_prof="SS",
    )

    print("next: NT")
    temp_profile_NT = accretion_disk_temperature(
        radii_in_meters,
        min_radius_in_meters,
        mass_in_m_sun,
        eddington_ratio,
        beta=0,
        corona_height=6,
        albedo=1,
        eta_x_rays=0.1,
        generic_beta=False,
        disk_acc=None,
        efficiency=0.1,
        spin=0,
        visc_temp_prof="NT",
    )

    print("next: SS + wind")
    temp_profile_SS_wind = accretion_disk_temperature(
        radii_in_meters,
        min_radius_in_meters,
        mass_in_m_sun * const.M_sun.to(u.kg),
        eddington_ratio,
        beta=0.72,
        corona_height=6,
        albedo=0,
        eta_x_rays=0.1,
        generic_beta=True,
        disk_acc=None,
        efficiency=0.1,
        spin=0,
        visc_temp_prof="SS",
    )

    print("next: rejected, so SS + wind")
    temp_profile_rejected_profile = accretion_disk_temperature(
        radii_in_meters * u.m,
        min_radius_in_meters / 1000 * u.km,
        mass_in_m_sun * const.M_sun.to(u.kg),
        eddington_ratio,
        beta=0.72,
        corona_height=6,
        albedo=0,
        eta_x_rays=0.1,
        generic_beta=True,
        disk_acc=None,
        efficiency=0.1,
        spin=0,
        visc_temp_prof="Not_a_profile",
    )
    print("this one below")

    temp_profile_little_accreted_mass = accretion_disk_temperature(
        radii_in_meters * u.m,
        min_radius_in_meters / 1000 * u.km,
        mass_in_m_sun * const.M_sun.to(u.g),
        eddington_ratio,
        disk_acc=0.001 * const.M_sun.to(u.kg) / u.yr,
        efficiency=0.1,
        spin=0,
        visc_temp_prof="SS",
    )

    temp_profile_another_little_accreted_mass = accretion_disk_temperature(
        radii_in_meters * u.m,
        min_radius_in_meters / 1000 * u.km,
        mass_in_m_sun * const.M_sun.to(u.g),
        eddington_ratio,
        disk_acc=0.001 * (const.M_sun.to(u.kg) / u.yr).to(u.kg / u.s).value,
        efficiency=0.1,
        spin=0,
        visc_temp_prof="SS",
    )

    # check that they're all different
    assert abs(sum(temp_profile_SS - temp_profile_NT)) != 0
    assert abs(sum(temp_profile_SS - temp_profile_SS_wind)) != 0

    # check that the rejected profile defaulted to a SS profile
    assert abs(sum(temp_profile_SS_wind - temp_profile_rejected_profile)) == 0

    assert np.sum(temp_profile_little_accreted_mass) < np.sum(temp_profile_SS)
    assert np.sum(temp_profile_little_accreted_mass) != 0

    # check that at large radii, SS and NT profiles converge
    large_radius = 10**5 * grav_rad
    temperature_difference_tolerance = 10  # Kelvins
    temp_large_radius_SS = accretion_disk_temperature(
        large_radius, min_radius_in_meters, mass_in_m_sun, eddington_ratio
    )
    temp_large_radius_NT = accretion_disk_temperature(
        large_radius,
        min_radius_in_meters,
        mass_in_m_sun,
        eddington_ratio,
        visc_temp_prof="NT",
    )
    assert (
        abs(temp_large_radius_SS - temp_large_radius_NT)
        < temperature_difference_tolerance
    )


def test_planck_law():
    temp_1 = 100
    temp_2 = 500
    wavelength_1 = 100
    wavelength_2 = 500

    radiance_1_1 = planck_law(temp_1, wavelength_1)
    radiance_1_2 = planck_law(temp_1, wavelength_2 * u.nm)
    radiance_2_1 = planck_law(temp_2, wavelength_2)

    # increasing wavelength and temperature should increase the radiance
    # for relatively low temps and short wavelengths.
    assert radiance_1_1 < radiance_1_2
    assert radiance_1_1 < radiance_2_1

    # giving a minimal temperature should provide minimal radiance
    temp_mini = 0.0001
    radiance_mini = planck_law(temp_mini, wavelength_1)
    tolerance = radiance_1_1 / 10**5
    assert radiance_mini == 0


def test_planck_law_derivative():
    temp_1 = 100
    temp_2 = 500
    wavelength = 100

    delta_radiance_1 = planck_law_derivative(temp_1, wavelength)
    delta_radiance_2 = planck_law_derivative(temp_2, wavelength)

    # At higher temperature, the radiance should be increasing faster
    # with respect to temp
    assert delta_radiance_2 > delta_radiance_1


def test_calculate_gravitational_radius():
    mass_1 = 10
    mass_2 = 20
    mass_astropy = 10 * const.M_sun
    mass_u_kg = 10 * u.kg

    grav_rad_1 = calculate_gravitational_radius(mass_1)
    grav_rad_2 = calculate_gravitational_radius(mass_2)
    grav_rad_astropy = calculate_gravitational_radius(mass_astropy)
    grav_rad_kg = calculate_gravitational_radius(mass_u_kg)

    assert grav_rad_1 == grav_rad_2 / 2
    assert grav_rad_1 == grav_rad_astropy
    assert grav_rad_1 > grav_rad_kg


def test_calculate_angular_diameter_distance():
    redshift = 0.1
    OmM = 0.3
    H0 = 70

    ang_diam_dist = calculate_angular_diameter_distance(redshift, OmM=OmM, H0=H0)
    astropy_cosmo = FlatLambdaCDM(H0, OmM)
    ang_diam_dist_astropy = (
        astropy_cosmo.angular_diameter_distance(redshift).to(u.m).value
    )

    # set tolerance to 0.5% due to quadrature integration and
    # rounding of constants
    tolerance = ang_diam_dist_astropy / 500

    assert abs(ang_diam_dist - ang_diam_dist_astropy) < tolerance


def test_calculate_angular_diameter_distance_difference():
    redshift_lens = 0.1
    redshift_source = 0.5
    OmM = 0.3
    H0 = 70

    ang_diam_dist_lens = calculate_angular_diameter_distance(
        redshift_lens, OmM=OmM, H0=H0
    )
    ang_diam_dist_source = calculate_angular_diameter_distance(
        redshift_source, OmM=OmM, H0=H0
    )

    ang_diam_dist_diff = calculate_angular_diameter_distance_difference(
        redshift_lens, redshift_source, OmM=OmM, H0=H0
    )

    ang_diam_dist_diff_reversed = calculate_angular_diameter_distance_difference(
        redshift_source, redshift_lens, OmM=OmM, H0=H0
    )

    # Check that redshift misordering doesn't cause issues
    assert ang_diam_dist_diff == ang_diam_dist_diff_reversed

    # Check that ADDD != ADD(source) - ADD(lens)
    assert ang_diam_dist_diff != (ang_diam_dist_source - ang_diam_dist_lens)

    # for extremely small redshifts, the ADDD should converge to ADD(source)

    redshift_lens_tiny = 10 ** (-10)
    redshift_source = 0.5

    ang_diam_dist_diff_tiny = calculate_angular_diameter_distance_difference(
        redshift_lens_tiny, redshift_source, OmM=OmM, H0=H0
    )
    ang_diam_dist_tiny = calculate_angular_diameter_distance(
        redshift_lens_tiny, OmM=OmM, H0=H0
    )
    # define a small tolerance
    tolerance = ang_diam_dist_source / 10**8
    assert (
        abs(ang_diam_dist_source - (ang_diam_dist_diff_tiny + ang_diam_dist_tiny))
        < tolerance
    )


def test_calculate_luminosity_distance():
    redshift = 0.1
    OmM = 0.3
    H0 = 70

    lum_dist = calculate_luminosity_distance(redshift, OmM=OmM, H0=H0)

    astropy_cosmo = FlatLambdaCDM(H0, OmM)
    lum_dist_astropy = astropy_cosmo.luminosity_distance(redshift).to(u.m).value

    # define a small tolerance due to parameter rounding
    tolerance = lum_dist_astropy / 500
    assert abs(lum_dist_astropy - lum_dist) < tolerance


def test_calculate_angular_einstein_radius():
    redshift_lens = 1.0
    redshift_source = 2.0
    avg_microlens_mass = 0.3 * const.M_sun.to(u.kg)
    OmM = 0.3
    H0 = 70
    test_distance_source = 300 * u.Mpc
    test_distance_lens = 100 * u.Mpc

    star_ang_ein_rad = calculate_angular_einstein_radius(
        redshift_lens,
        redshift_source,
        mean_microlens_mass_in_kg=avg_microlens_mass,
        OmM=OmM,
        H0=H0,
    )

    human_mass = 75 * u.kg
    human_ang_ein_rad = calculate_angular_einstein_radius(
        redshift_lens,
        redshift_source,
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
    )

    assert star_ang_ein_rad > human_ang_ein_rad
    assert isinstance(star_ang_ein_rad, float)

    # make sure the code switches the distances in case they are input backwards
    nearby_einstein_radius = calculate_angular_einstein_radius(
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
        D_lens=test_distance_source,
        D_source=test_distance_lens,
    )
    nearby_einstein_radius_2 = calculate_angular_einstein_radius(
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
        D_lens=test_distance_lens,
        D_source=test_distance_source,
    )
    assert nearby_einstein_radius == nearby_einstein_radius_2


def test_calculate_einstein_radius_in_meters():
    redshift_lens = 1.0
    redshift_source = 2.0
    avg_microlens_mass = 0.3 * const.M_sun.to(u.kg)
    OmM = 0.3
    H0 = 70
    test_distance_source = 300 * u.Mpc
    test_distance_lens = 100 * u.Mpc

    star_ein_rad = calculate_einstein_radius_in_meters(
        redshift_lens,
        redshift_source,
        mean_microlens_mass_in_kg=avg_microlens_mass,
        OmM=OmM,
        H0=H0,
    )

    human_mass = 75 * u.kg
    human_ein_rad = calculate_einstein_radius_in_meters(
        redshift_lens,
        redshift_source,
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
    )

    assert star_ein_rad > human_ein_rad
    # Test that we removed the astropy units
    assert isinstance(star_ein_rad, (float, int))

    nearby_einstein_radius = calculate_einstein_radius_in_meters(
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
        D_lens=test_distance_source,
        D_source=test_distance_lens,
    )
    nearby_einstein_radius_2 = calculate_einstein_radius_in_meters(
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
        D_lens=test_distance_lens,
        D_source=test_distance_source,
    )
    assert nearby_einstein_radius == nearby_einstein_radius_2

    backwards_einstein_radius = calculate_einstein_radius_in_meters(
        redshift_lens=redshift_source,
        redshift_source=redshift_lens,
        mean_microlens_mass_in_kg=human_mass,
        OmM=OmM,
        H0=H0,
    )
    assert backwards_einstein_radius == human_ein_rad


def test_pull_value_from_grid():
    x_values = np.linspace(1, 10, 10)
    y_values = np.linspace(2, 20, 10)

    # Note that arrays are created internally as (y, x)
    X, Y = np.meshgrid(x_values, y_values)

    test_grid = Y + X
    assert test_grid[1, 0] == 5
    assert test_grid[0, 1] == 4

    # explicitly pull values off grid
    assert test_grid[0, 0] == 3
    assert pull_value_from_grid(test_grid, 0, 0) == 3

    # test max value
    assert test_grid[-1, -1] == pull_value_from_grid(test_grid, 9, 9)

    # test small changes follow expected interpolation
    assert pull_value_from_grid(test_grid, 0, 0.5) == 4.0
    assert pull_value_from_grid(test_grid, 0.5, 0) == 3.5

    # show dx < dy
    assert pull_value_from_grid(test_grid, 5, 5.2) > pull_value_from_grid(
        test_grid, 5.2, 5
    )

    # test multiple values pulled at once
    pull_values_x = np.linspace(3, 5, 10)
    pull_values_y = np.linspace(5, 2, 10)
    assert len(pull_values_x) == len(pull_values_y)
    output_values = pull_value_from_grid(
        test_grid,
        pull_values_x,
        pull_values_y,
    )

    # show that these are equivalent
    assert len(output_values) == len(pull_values_y)
    for jj in range(len(pull_values_x)):
        current_value = pull_value_from_grid(
            test_grid, pull_values_x[jj], pull_values_y[jj]
        )
        assert output_values[jj] == current_value


def test_convert_1d_array_to_2d_array():
    test_list = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    repacked_array = convert_1d_array_to_2d_array(test_list)
    assert np.shape(repacked_array) == (3, 3)
    assert repacked_array[0, 0] == 1
    assert repacked_array[-1, -1] == 9
    assert repacked_array[0, 1] == 2
    assert repacked_array[1, 0] == 4


def test_conversions_between_cartesian_and_polar():
    x1 = 1
    y1 = 2

    hyp1 = (x1**2 + y1**2) ** 0.5

    # we want phi = 0 pointing along the -y axis, to the observer if
    # the plane is inclined. Need to test using a different numpy function.

    expected_angle = np.arcsin(y1 / hyp1) + np.pi / 2

    x2 = 2
    y2 = 4

    r1, azi1 = convert_cartesian_to_polar(x1, y1)
    r2, azi2 = convert_cartesian_to_polar(x2, y2)

    np.testing.assert_almost_equal(azi1, expected_angle)

    assert azi2 == azi1
    assert r1 == (x1**2 + y1**2) ** 0.5
    assert r2 == (x2**2 + y2**2) ** 0.5

    # show this is invertible
    x_out_1, y_out_1 = convert_polar_to_cartesian(r1, azi1)
    x_out_2, y_out_2 = convert_polar_to_cartesian(r2, azi2)

    # internal rounding makes these off by ~10^-8
    np.testing.assert_almost_equal(x1, x_out_1)
    np.testing.assert_almost_equal(x2, x_out_2)
    np.testing.assert_almost_equal(y1, y_out_1)
    np.testing.assert_almost_equal(y2, y_out_2)

    _, x_axis = convert_cartesian_to_polar(1, 0)
    _, y_axis = convert_cartesian_to_polar(0, 1)
    _, neg_x_axis = convert_cartesian_to_polar(-1, 0)
    _, neg_y_axis = convert_cartesian_to_polar(0, -1)

    np.testing.assert_almost_equal(x_axis, np.pi / 2)
    np.testing.assert_almost_equal(y_axis, np.pi)
    np.testing.assert_almost_equal(neg_x_axis, 3 * np.pi / 2)
    np.testing.assert_almost_equal(neg_y_axis, 0)


def test_perform_microlensing_convolution():
    magnification_array_identity = np.ones((100, 100))
    x_ax = 5 - abs(5 - np.linspace(1, 10, 10))
    flux_x, flux_y = np.meshgrid(x_ax, x_ax)
    sample_flux_map = (flux_x**2 + flux_y**2) ** 0.5

    redshift_l = 0.5
    redshift_s = 2.0
    relative_orientation_1 = 0
    relative_orientation_1a = 360
    relative_orientation_2 = 180

    convolution_1, px_shift_1 = perform_microlensing_convolution(
        magnification_array_identity,
        sample_flux_map,
        redshift_l,
        redshift_s,
        relative_orientation=relative_orientation_1,
    )
    convolution_1a, px_shift_1a = perform_microlensing_convolution(
        magnification_array_identity,
        sample_flux_map,
        redshift_l,
        redshift_s,
        relative_orientation=relative_orientation_1a,
    )
    convolution_2, px_shift_2 = perform_microlensing_convolution(
        magnification_array_identity,
        sample_flux_map,
        redshift_l,
        redshift_s,
        relative_orientation=relative_orientation_2,
    )
    convolution_2b, px_shift_2b = perform_microlensing_convolution(
        magnification_array_identity,
        sample_flux_map,
        redshift_l,
        redshift_s,
        relative_orientation=None,
    )
    assert px_shift_2 == px_shift_2b

    # test that we conserve flux after accounting for rescaling of pixels
    value = convolution_1[3, 2]
    npt.assert_almost_equal(value, np.sum(sample_flux_map))

    # test the identity convolution is constant
    npt.assert_almost_equal(convolution_1a[2, 5], convolution_1a[-4, -1])

    # test that rotation of 180, 360 deg is identity for symmetric source
    npt.assert_almost_equal(np.sum(convolution_1), np.sum(convolution_1a))
    npt.assert_almost_equal(np.sum(convolution_1), np.sum(convolution_2))

    assert px_shift_1 == px_shift_1a
    assert px_shift_1 == px_shift_2

    # by defining 2 point sources, the convolution should only have
    # one point source, spread by the pixel shift
    mag_array_point = np.zeros(np.shape(magnification_array_identity))
    mag_array_point[75, 25] = 1

    convolution_2_points, px_shift = perform_microlensing_convolution(
        mag_array_point,
        mag_array_point,
        redshift_l,
        redshift_s,
        relative_orientation=relative_orientation_1,
    )

    total_value = np.sum(convolution_2_points)
    npt.assert_almost_equal(total_value, 1)


def test_extract_light_curve():
    ones_array = np.ones((100, 100))
    fake_flux_distribution = [
        [0, 0, 1, 0, 0],
        [0, 1, 2, 1, 0],
        [1, 2, 3, 2, 1],
        [0, 1, 2, 1, 0],
        [0, 0, 1, 0, 0],
    ]
    effective_vel = 10000
    effective_vel_too_fast = 10**20
    light_curve_length = 5
    # these parameters make the light curve 10 pixels long
    pixel_size = calculate_einstein_radius_in_meters(
        redshift_lens=1.0,
        redshift_source=2.0,
        mean_microlens_mass_in_kg=0.3 * const.M_sun.to(u.kg),
        OmM=0.3,
        H0=70,
    )

    convolution_1, px_shift_1 = perform_microlensing_convolution(
        ones_array,
        fake_flux_distribution,
        1.0,
        2.0,
        mean_microlens_mass_in_kg=0.3 * const.M_sun.to(u.kg),
        number_of_microlens_einstein_radii=1,
        number_of_smbh_gravitational_radii=20,
    )

    light_curve = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
    )

    light_curve2 = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel * u.km / u.s,
        light_curve_length * u.yr,
        px_shift_1,
    )

    for tt in range(len(light_curve)):
        npt.assert_almost_equal(light_curve[tt], np.asarray(light_curve).mean())

    original_total_integrated_flux = np.sum(fake_flux_distribution)
    convolved_total_flux = light_curve[0]

    npt.assert_almost_equal(original_total_integrated_flux, convolved_total_flux)

    # check we can pull 1000 light curves from this convolution
    # to show it never extends beyond the edge
    for jj in range(1000):
        extract_light_curve(
            convolution_1,
            pixel_size,
            effective_vel,
            light_curve_length,
            px_shift_1,
        )

    light_curve, x_pos, y_pos = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        return_track_coords=True,
    )

    # show the light curve is drawn from multiple positions, which
    # we can almost guarantee because rng will never pick exactly
    # n*90 deg for int(n).
    assert x_pos[0] != x_pos[-1]
    assert y_pos[0] != y_pos[-1]

    # show we can pick a specific light curve
    # note that coordinates on an array are in signature (y, x)
    light_curve, x_pos, y_pos = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        x_start_position=30,
        y_start_position=40,
        phi_travel_direction=90,
        return_track_coords=True,
    )

    assert x_pos[0] == x_pos[-1]
    assert y_pos[0] != y_pos[-1]

    light_curve, x_pos, y_pos = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        x_start_position=30,
        y_start_position=40,
        phi_travel_direction=180,
        return_track_coords=True,
    )

    assert x_pos[0] != x_pos[-1]
    assert y_pos[0] == y_pos[-1]

    # show we can get the average value for many cases
    # e.g. source too large (high pixel shift)
    light_curve_avg_1 = extract_light_curve(
        convolution_1, pixel_size, effective_vel, light_curve_length, 300
    )
    # e.g. light curve too long
    light_curve_avg_2 = extract_light_curve(
        convolution_1, pixel_size, effective_vel * 10e8, light_curve_length, px_shift_1
    )
    # e.g. direction chosen to extend beyond the convolution's boundaries
    light_curve_avg_3 = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        x_start_position=90,
        y_start_position=90,
        phi_travel_direction=0,
    )
    light_curve_avg_4 = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        x_start_position=90,
        y_start_position=-90,
        phi_travel_direction=0,
    )
    light_curve_avg_5 = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        x_start_position=-90,
        y_start_position=90,
        phi_travel_direction=0,
    )
    light_curve_avg_6 = extract_light_curve(
        convolution_1,
        pixel_size,
        effective_vel,
        light_curve_length,
        px_shift_1,
        x_start_position=99,
        y_start_position=99,
        phi_travel_direction=27,
    )

    assert light_curve_avg_1 == light_curve_avg_2
    assert light_curve_avg_3 == light_curve_avg_2
    assert light_curve_avg_1 == light_curve_avg_4
    assert light_curve_avg_1 == light_curve_avg_5
    assert light_curve_avg_1 == light_curve_avg_6


def test_calculate_time_lag_array():
    # point source tests

    radii_array = 0
    phi_array = 0
    inclination_angle = 0
    corona_height = 10
    axis_offset_in_gravitational_radii = 0
    angle_offset_in_degrees = 0
    height_array = 0

    time_lag_1 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
    )

    assert time_lag_1 == (2 * corona_height)

    corona_height = 0
    time_lag_2 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
    )

    assert time_lag_2 == (2 * corona_height)

    # test manual construction of radii and phi arrays
    root2 = np.sqrt(2)

    radii_array = np.asarray([[root2, 1, root2], [1, 0, 1], [root2, 1, root2]])

    phi_array = np.asarray(
        [
            [5 * np.pi / 4, np.pi, 3 * np.pi / 4],
            [6 * np.pi / 4, 0, 2 * np.pi / 4],
            [7 * np.pi / 4, 0, 1 * np.pi / 4],
        ]
    )

    height_array = None
    corona_height = 10

    time_lag_3 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
    )

    assert time_lag_3[0, 0] == time_lag_3[-1, -1]
    assert time_lag_3[0, 0] == time_lag_3[0, -1]
    assert time_lag_3[1, 0] == time_lag_3[1, -1]
    assert time_lag_3[0, 1] == time_lag_3[-1, 1]
    assert time_lag_3[1, 1] == 2 * corona_height

    # test off axis time lags
    axis_offset_in_gravitational_radii = 1

    time_lag_4 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
    )

    assert np.min(time_lag_4) == 2 * corona_height

    # rotate lamppost to other side of disk
    angle_offset_in_degrees = 180
    time_lag_5 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
    )

    assert np.min(time_lag_5) == 2 * corona_height
    assert time_lag_4[1, -1] < time_lag_5[1, -1]
    assert time_lag_4[0, 1] == time_lag_5[0, 1]
    assert time_lag_4[1, 0] > time_lag_5[1, 0]

    # test with heights included (non-flat accretion disk)
    axis_offset_in_gravitational_radii = 0
    height_array = radii_array
    corona_height = 10

    time_lag_6 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
        height_array=height_array,
    )

    # with disk flaring, center time lag should be longest lag.
    # argmax of a 2d array flattens the array, so index 4 is center.
    assert np.argmax(time_lag_6) == 4
    assert time_lag_6[1, 1] == 2 * corona_height

    # compare with other points on isodelay surface
    assert time_lag_6[1, 0] == time_lag_6[1, -1]
    assert time_lag_6[1, 0] == time_lag_6[1, 0]
    assert time_lag_6[1, 0] == time_lag_6[-1, 1]

    # explicitly calculate delay for one point
    assert (
        time_lag_6[1, 1]
        == ((corona_height - height_array[1, 1]) ** 2 + radii_array[1, 1] ** 2) ** 0.5
        + corona_height
    )

    # test inclined disk
    inclination_angle = 35
    height_array = None

    time_lag_7 = calculate_time_lag_array(
        radii_array,
        phi_array,
        inclination_angle,
        corona_height,
    )

    assert time_lag_7[1, 0] == time_lag_7[1, -1]
    assert time_lag_7[-1, 1] < time_lag_7[0, 1]


def test_calculate_geometric_disk_factor():
    temp_array = np.ones((100, 100))
    x_vals = abs(np.linspace(1, 100, 100) - 50)
    X, Y = np.meshgrid(x_vals, x_vals)
    radii_map, phi_map = convert_cartesian_to_polar(X, Y)
    smbh_mass_exponent = 8.0
    corona_height = 10
    height_array = radii_map / 10

    geo_disk_factor_array = calculate_geometric_disk_factor(
        temp_array,
        radii_map,
        phi_map,
        smbh_mass_exponent,
        corona_height,
        height_array=height_array,
    )

    assert geo_disk_factor_array[50, 50] > geo_disk_factor_array[0, 50]

    albedo_array = 0 + 0.1 * radii_map / np.max(radii_map)

    geo_disk_factor_array_albedos = calculate_geometric_disk_factor(
        temp_array,
        radii_map,
        phi_map,
        smbh_mass_exponent,
        corona_height,
        height_array=height_array,
        albedo_array=albedo_array,
    )
    geo_disk_factor_array_const_albedos = calculate_geometric_disk_factor(
        temp_array,
        radii_map,
        phi_map,
        smbh_mass_exponent,
        corona_height,
        height_array=height_array,
        albedo_array=0.75,
    )

    # since higher albedo means less absorption, every position (except [50, 50])
    # will be less than before
    test_x = 35
    test_y = 73
    assert (
        geo_disk_factor_array_albedos[test_x, test_y]
        < geo_disk_factor_array[test_x, test_y]
    )

    # show we can input off-axis source
    axis_offset_in_gravitational_radii = 10
    angle_offset_in_degrees = 47

    calculate_geometric_disk_factor(
        temp_array,
        radii_map,
        phi_map,
        smbh_mass_exponent,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
        albedo_array=albedo_array,
    )


def test_calculate_dt_dlx():

    temp_array = 1000 * np.ones((100, 100))
    # must handle zeros in temp profile
    temp_array[47:53, 47:53] = 0

    x_vals = abs(np.linspace(1, 100, 100) - 50)
    X, Y = np.meshgrid(x_vals, x_vals)
    radii_array, phi_array = convert_cartesian_to_polar(X, Y)
    smbh_mass_exponent = 8.0
    corona_height = 10
    height_array = radii_array / 10

    axis_offset_in_gravitational_radii = 0
    angle_offset_in_degrees = 0
    albedo_array = None

    dt_dlx = calculate_dt_dlx(
        temp_array,
        radii_array,
        phi_array,
        smbh_mass_exponent,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
        albedo_array=albedo_array,
    )

    # we shouldn't be injecting a massive amount of energy
    # also show that nans are removed
    assert np.sum(dt_dlx) < 10**3


def test_construct_accretion_disk_transfer_function():

    test_wavelength_1 = 100  # nm
    test_wavelength_2 = 300
    smbh_mass_exponent = 8.0
    corona_height = 10
    inclination_angle = 30
    axis_offset_in_gravitational_radii = 0
    angle_offset_in_degrees = 0
    height_array = None
    albedo_array = None

    # to test reasonable transfer functions, we must make a reasonable disk
    disk_dict = create_maps(smbh_mass_exponent, 0, 500, inclination_angle, 500)

    temp_array = disk_dict["temp_array"]
    radii_array = disk_dict["radii_array"]
    phi_array = disk_dict["phi_array"]
    g_array = disk_dict["g_array"]
    inclination_angle = disk_dict["inclination_angle"]
    smbh_mass_exponent = disk_dict["smbh_mass_exp"]
    corona_height = disk_dict["corona_height"]

    transfer_function_test_1 = construct_accretion_disk_transfer_function(
        test_wavelength_1,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
        albedo_array=albedo_array,
    )

    # show it was normalized
    npt.assert_almost_equal(np.sum(transfer_function_test_1), 1)
    # there should be an initial time lag
    assert transfer_function_test_1[0] == 0
    assert transfer_function_test_1[5] == 0
    assert transfer_function_test_1[8] == 0

    # the same disk at a redder wavelength should peak at later time lag
    transfer_function_test_2 = construct_accretion_disk_transfer_function(
        test_wavelength_2,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
        albedo_array=albedo_array,
    )

    tau_axis = np.linspace(
        0, len(transfer_function_test_1) - 1, len(transfer_function_test_1)
    )
    mean_lag_1 = np.sum(tau_axis * transfer_function_test_1)
    mean_lag_2 = np.sum(tau_axis * transfer_function_test_2)
    assert mean_lag_1 < mean_lag_2


def test_calculate_microlensed_transfer_function():

    test_wavelength = 300  # nm
    smbh_mass_exponent = 8.0
    corona_height = 10
    inclination_angle = 30
    axis_offset_in_gravitational_radii = 0
    angle_offset_in_degrees = 0
    height_array = None
    albedo_array = None

    # to test reasonable transfer functions, we must make a reasonable disk
    disk_dict = create_maps(smbh_mass_exponent, 0, 500, inclination_angle, 500)

    temp_array = disk_dict["temp_array"]
    radii_array = disk_dict["radii_array"]
    phi_array = disk_dict["phi_array"]
    g_array = disk_dict["g_array"]
    inclination_angle = disk_dict["inclination_angle"]
    smbh_mass_exponent = disk_dict["smbh_mass_exp"]
    corona_height = disk_dict["corona_height"]

    transfer_function_test_no_ml = construct_accretion_disk_transfer_function(
        test_wavelength,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
        angle_offset_in_degrees=angle_offset_in_degrees,
        height_array=height_array,
        albedo_array=albedo_array,
    )

    magnification_array_identity = np.ones((1000, 1000))

    redshift_l = 0.5
    redshift_s = 2.0
    relative_orientation_1 = 0
    mean_microlens_mass_in_kg = 0.3 * const.M_sun.to(u.kg)
    number_of_microlens_einstein_radii = 5

    orientation_1 = 0
    orientation_2 = 45

    # show we can reclaim the same transfer function when identity is used
    transfer_function_test_id_ml = calculate_microlensed_transfer_function(
        magnification_array_identity,
        redshift_l,
        redshift_s,
        test_wavelength,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        mean_microlens_mass_in_kg=mean_microlens_mass_in_kg,
        number_of_microlens_einstein_radii=number_of_microlens_einstein_radii,
        number_of_smbh_gravitational_radii=1000,
        relative_orientation=orientation_1,
    )

    transfer_function_test_id_ml_rotate = calculate_microlensed_transfer_function(
        magnification_array_identity,
        redshift_l,
        redshift_s,
        test_wavelength,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        mean_microlens_mass_in_kg=mean_microlens_mass_in_kg,
        number_of_microlens_einstein_radii=number_of_microlens_einstein_radii,
        number_of_smbh_gravitational_radii=1000,
        relative_orientation=orientation_2,
    )

    data_to_construct_transfer_function = calculate_microlensed_transfer_function(
        magnification_array_identity,
        redshift_l,
        redshift_s,
        test_wavelength,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        mean_microlens_mass_in_kg=mean_microlens_mass_in_kg,
        number_of_microlens_einstein_radii=number_of_microlens_einstein_radii,
        number_of_smbh_gravitational_radii=1000,
        relative_orientation=orientation_2,
        return_response_array_and_lags=True,
    )
    descaled_data = calculate_microlensed_transfer_function(
        magnification_array_identity,
        redshift_l,
        redshift_s,
        test_wavelength,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        mean_microlens_mass_in_kg=mean_microlens_mass_in_kg,
        number_of_microlens_einstein_radii=number_of_microlens_einstein_radii,
        number_of_smbh_gravitational_radii=1000,
        relative_orientation=orientation_2,
        return_descaled_response_array_and_lags=True,
    )

    tau_axis_ml = np.linspace(
        0, len(transfer_function_test_id_ml) - 1, len(transfer_function_test_id_ml)
    )

    tau_axis_no_ml = np.linspace(
        0, len(transfer_function_test_no_ml) - 1, len(transfer_function_test_no_ml)
    )
    mean_tau_no_ml = np.sum(transfer_function_test_no_ml * tau_axis_no_ml) / np.sum(
        transfer_function_test_no_ml
    )
    mean_tau_id_ml = np.sum(transfer_function_test_id_ml * tau_axis_ml) / np.sum(
        transfer_function_test_id_ml
    )
    mean_tau_id_ml_rotated = np.sum(
        transfer_function_test_id_ml_rotate * tau_axis_ml
    ) / np.sum(transfer_function_test_id_ml_rotate)

    # note there is some rounding when rescale and rotate are used and inverted
    tolerance = mean_tau_no_ml / 20
    assert abs(mean_tau_no_ml - mean_tau_id_ml) <= tolerance
    current_diff = mean_tau_no_ml - mean_tau_id_ml_rotated
    assert abs(current_diff) <= tolerance

    assert len(data_to_construct_transfer_function) == 4
    assert np.size(data_to_construct_transfer_function[0]) == np.size(
        data_to_construct_transfer_function[1]
    )
    assert len(descaled_data) == 4
    assert np.size(descaled_data[0]) == np.size(descaled_data[1])

    # test the magnification crop
    crop_data = calculate_microlensed_transfer_function(
        magnification_array_identity,
        redshift_l,
        redshift_s,
        test_wavelength,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        inclination_angle,
        smbh_mass_exponent,
        corona_height,
        mean_microlens_mass_in_kg=mean_microlens_mass_in_kg,
        number_of_microlens_einstein_radii=number_of_microlens_einstein_radii,
        number_of_smbh_gravitational_radii=1000,
        relative_orientation=orientation_2,
        return_magnification_map_crop=True,
    )


def test_generate_drw_signal():
    maximum_time = 2000
    time_step = 2
    sf_inf_1 = 100
    tau_drw_1 = 100
    random_seed = 17
    sf_inf_2 = 400
    tau_drw_2 = 10

    drw_1 = generate_drw_signal(
        maximum_time, time_step, sf_inf_1, tau_drw_1, random_seed=random_seed
    )

    drw_2 = generate_drw_signal(
        maximum_time, time_step, sf_inf_2, tau_drw_2, random_seed=random_seed
    )

    assert len(drw_1) == len(drw_2)
    npt.assert_almost_equal(np.mean(drw_1), np.mean(drw_2))

    random_point_1 = np.random.randint(500)
    random_point_2 = np.random.randint(500)

    if random_point_1 == random_point_2:  # pragma: no cover
        random_point_2 -= 1

    assert drw_1[random_point_1] != drw_1[random_point_2]
    assert drw_2[random_point_1] != drw_2[random_point_2]


def test_generate_signal_from_psd():
    length_of_light_curve = 1000
    frequencies = np.linspace(
        1 / length_of_light_curve, 1 / 4, length_of_light_curve + 1
    )
    psd_flat_spectrum = np.ones(np.shape(frequencies))
    psd_power_law = frequencies ** (-2)
    random_seed = 33

    time_ax, lc_flat_spectrum = generate_signal_from_psd(
        length_of_light_curve,
        psd_flat_spectrum,
        frequencies,
        random_seed=random_seed,
    )

    time_ax, lc_power_spectrum = generate_signal_from_psd(
        length_of_light_curve,
        psd_power_law,
        frequencies,
        random_seed=random_seed,
    )

    assert len(lc_flat_spectrum) == len(lc_power_spectrum)
    assert lc_flat_spectrum[3] != lc_flat_spectrum[0]


def test_generate_snapshots_of_radiation_pattern():

    test_wavelength = 700  # nm
    smbh_mass_exponent = 8.0
    corona_height = 10
    inclination_angle = 30

    # to test radiation patterns, we must make a reasonable disk
    disk_dict = create_maps(smbh_mass_exponent, 0, 1000, inclination_angle, 1000)

    temp_array = disk_dict["temp_array"]
    radii_array = disk_dict["radii_array"]
    phi_array = disk_dict["phi_array"]
    g_array = disk_dict["g_array"]
    inclination_angle = disk_dict["inclination_angle"]
    smbh_mass_exponent = disk_dict["smbh_mass_exp"]
    corona_height = disk_dict["corona_height"]

    static_disk_radiation = planck_law(temp_array, test_wavelength)

    snapshot_list = [0, 10, 100, 500]
    maximum_time = 1000
    time_step = 1
    sf_inf = 50
    tau_drw = 100
    random_seed = 40

    driving_signal_fractional_strength_half = 0.5
    driving_signal_fractional_strength_zero = 0.0
    driving_signal_fractional_strength_all = 1.0

    drw_signal = generate_drw_signal(
        maximum_time, time_step, sf_inf, tau_drw, random_seed=random_seed
    )

    snapshots_half = generate_snapshots_of_radiation_pattern(
        test_wavelength,
        snapshot_list,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        smbh_mass_exponent,
        drw_signal,
        driving_signal_fractional_strength_half,
        corona_height,
        inclination_angle,
    )

    snapshots_zero = generate_snapshots_of_radiation_pattern(
        test_wavelength,
        snapshot_list,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        smbh_mass_exponent,
        drw_signal,
        driving_signal_fractional_strength_zero,
        corona_height,
        inclination_angle,
    )

    snapshots_all = generate_snapshots_of_radiation_pattern(
        test_wavelength,
        snapshot_list,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        smbh_mass_exponent,
        drw_signal,
        driving_signal_fractional_strength_all,
        corona_height,
        inclination_angle,
    )
    snapshots_short_driving_signal = generate_snapshots_of_radiation_pattern(
        test_wavelength,
        snapshot_list,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        smbh_mass_exponent,
        drw_signal[:5],
        driving_signal_fractional_strength_all,
        corona_height,
        inclination_angle,
    )

    assert np.shape(snapshots_half) == np.shape(snapshots_zero)
    assert np.shape(snapshots_half) == np.shape(snapshots_all)

    for jj in range(len(snapshot_list)):
        # use abs to show each pixel doesn't deviate from static case
        assert np.sum(abs(snapshots_zero[jj] - static_disk_radiation)) == 0
        # like before, but show both variabile disks do deviate
        assert np.sum(abs(snapshots_half[jj] - static_disk_radiation)) > 0
        assert np.sum(abs(snapshots_all[jj] - static_disk_radiation)) > 0

    signal_increasing_power = np.linspace(1, 10000, 10000)

    snapshots_rising_power = generate_snapshots_of_radiation_pattern(
        test_wavelength,
        snapshot_list,
        temp_array,
        radii_array,
        phi_array,
        g_array,
        smbh_mass_exponent,
        signal_increasing_power,
        driving_signal_fractional_strength_all,
        corona_height,
        inclination_angle,
    )

    for jj in range(len(snapshot_list) - 1):
        assert np.sum(snapshots_rising_power[jj + 1] - snapshots_rising_power[jj]) > 0


def test_project_blr_to_source_plane():

    # create some relatively large density region
    # want 1000 in the R direction, 50 in the Z direction so the
    # loop over each slab doesn't take long

    max_radius = 500
    max_height = 500

    r_points = 20
    z_points = 10

    r_step = max_radius // r_points
    z_step = max_height // z_points

    test_density = np.ones((r_points, z_points))

    radial_vel_grid = np.zeros(np.shape(test_density))
    vertical_vel_grid = np.zeros(np.shape(test_density))
    vertical_vel_grid[:, 4:] = 0.3
    vertical_vel_grid[:, 8:] = 0.6

    radii = np.linspace(0, max_radius, r_points)
    heights = np.linspace(0, max_height, z_points)

    R, Z = np.meshgrid(radii, heights, indexing="ij")
    distance_grid = (R**2 + Z**2) ** 0.5
    weighting = np.exp(-((distance_grid - 500) ** 2))

    smbh_exp = 8.0
    inclination = 0

    with pytest.raises(AssertionError):
        project_blr_to_source_plane(
            test_density,
            vertical_vel_grid,
            radial_vel_grid,
            -3,
            smbh_exp,
            velocity_range=[-1, 1],
            weighting_grid=weighting,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        project_blr_to_source_plane(
            test_density,
            vertical_vel_grid,
            radial_vel_grid,
            100,
            smbh_exp,
            velocity_range=[-1, 1],
            weighting_grid=weighting,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        project_blr_to_source_plane(
            test_density,
            vertical_vel_grid.T,
            radial_vel_grid,
            inclination,
            smbh_exp,
            velocity_range=[-1, 1],
            weighting_grid=weighting,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        project_blr_to_source_plane(
            test_density,
            vertical_vel_grid,
            radial_vel_grid.T,
            inclination,
            smbh_exp,
            velocity_range=[-1, 1],
            weighting_grid=weighting,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        project_blr_to_source_plane(
            test_density,
            vertical_vel_grid,
            radial_vel_grid,
            inclination,
            smbh_exp,
            velocity_range=[-1, 1],
            weighting_grid=weighting.T,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )

    test_projection_1 = project_blr_to_source_plane(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination,
        smbh_exp,
        velocity_range=[-1, 1],
        weighting_grid=weighting,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )[0]

    # since a cylinder of constant density is being projected, a projected
    # row in the center will have a greater value than an edge row in all cases
    assert np.sum(test_projection_1[0]) < np.sum(test_projection_1[r_points])
    assert np.sum(test_projection_1[-1]) < np.sum(test_projection_1[r_points])
    assert np.sum(test_projection_1.T[0]) < np.sum(test_projection_1[r_points])
    assert np.sum(test_projection_1.T[-1]) < np.sum(test_projection_1[r_points])

    # if inclined, the edge row of values on the far side will be greater than
    # the zeros on the near side
    inclination = 20
    test_projection_2 = project_blr_to_source_plane(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination,
        smbh_exp,
        velocity_range=[-1, 1],
        weighting_grid=None,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )[0]

    # note the projection is enlarged w.r.t. the original source to capture
    # the whole projection
    projection_padding = np.size(test_projection_2, 0) // 2 - r_points

    assert np.sum(test_projection_2[projection_padding + 10]) < np.sum(
        test_projection_2[-(projection_padding + 10)]
    )
    # if transposed, both sides should remain the same
    assert np.sum(test_projection_2.T[projection_padding]) == np.sum(
        test_projection_2.T[-projection_padding]
    )
    assert np.sum(test_projection_2.T[(projection_padding + 1)]) == np.sum(
        test_projection_2.T[-(projection_padding + 2)]
    )

    # test that we can only select certain velocities (receeding in this case)
    inclination = 60
    velocity_range_receeding = [-1, -0.001]
    vertical_vel_grid = np.zeros(np.shape(test_density))

    test_projection_3_receeding = project_blr_to_source_plane(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination,
        smbh_exp,
        velocity_range=velocity_range_receeding,
        weighting_grid=weighting,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )[0]

    velocity_range_approaching = [0.001, 1]

    test_projection_3_approaching = project_blr_to_source_plane(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination,
        smbh_exp,
        velocity_range=velocity_range_approaching,
        weighting_grid=weighting,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )[0]

    projection_padding = np.size(test_projection_3_approaching, 0) // 2 - r_points
    middle = np.size(test_projection_3_approaching, 0) // 2

    # when inclined, the velocity is a mix of each direction.
    # in constructing this there was no radial or vertical velocity added.
    # therefore, the los velocity depends on phi coord.

    # test phi dependence on lhs and rhs of the black hole
    assert np.sum(test_projection_3_approaching[:, middle + 5]) < np.sum(
        test_projection_3_receeding[:, middle + 5]
    )
    assert np.sum(test_projection_3_approaching[:, middle - 5]) > np.sum(
        test_projection_3_receeding[:, middle - 5]
    )

    # test outflow viewed perfectly face on by projecting everything to approaching grid
    inclination = 0
    for jj in range(np.size(vertical_vel_grid, 1)):
        vertical_vel_grid[:, jj] = 0.5 * jj / np.size(vertical_vel_grid, 1)

    test_projection_4_receeding = project_blr_to_source_plane(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination,
        smbh_exp,
        velocity_range=velocity_range_receeding,
        weighting_grid=weighting,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )[0]
    test_projection_4_approaching = project_blr_to_source_plane(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination,
        smbh_exp,
        velocity_range=velocity_range_approaching,
        weighting_grid=weighting,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )[0]

    assert np.sum(test_projection_4_approaching) > 0
    assert np.sum(test_projection_4_receeding) == 0


def test_calculate_blr_transfer_function():

    max_radius = 100
    max_height = 100

    r_points = 20
    z_points = 10

    r_step = max_radius // r_points
    z_step = max_height // z_points

    test_density = np.ones((r_points, z_points))

    radial_vel_grid = np.zeros(np.shape(test_density))
    vertical_vel_grid = np.zeros(np.shape(test_density))
    vertical_vel_grid[:, 4:] = 0.3
    vertical_vel_grid[:, 8:] = 0.6

    inclination_angle = 0
    smbh_mass_exponent = 10

    with pytest.raises(AssertionError):
        calculate_blr_transfer_function(
            test_density,
            vertical_vel_grid,
            radial_vel_grid,
            -2,
            smbh_mass_exponent,
            velocity_range=[-1, 1],
            weighting_grid=None,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        calculate_blr_transfer_function(
            test_density,
            vertical_vel_grid,
            radial_vel_grid,
            50000,
            smbh_mass_exponent,
            velocity_range=[-1, 1],
            weighting_grid=None,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        calculate_blr_transfer_function(
            test_density,
            vertical_vel_grid.T,
            radial_vel_grid,
            -2,
            smbh_mass_exponent,
            velocity_range=[-1, 1],
            weighting_grid=None,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        calculate_blr_transfer_function(
            test_density,
            vertical_vel_grid,
            radial_vel_grid.T,
            -2,
            smbh_mass_exponent,
            velocity_range=[-1, 1],
            weighting_grid=None,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )
    with pytest.raises(AssertionError):
        calculate_blr_transfer_function(
            test_density,
            vertical_vel_grid,
            radial_vel_grid,
            -2,
            smbh_mass_exponent,
            velocity_range=[-1, 1],
            weighting_grid=test_density.T,
            radial_resolution=r_step,
            vertical_resolution=z_step,
        )

    face_on_blr_tf_1 = calculate_blr_transfer_function(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination_angle,
        smbh_mass_exponent,
        velocity_range=[-1, 1],
        weighting_grid=None,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )

    assert face_on_blr_tf_1.ndim == 1

    tau_ax = np.linspace(0, np.size(face_on_blr_tf_1) - 1, np.size(face_on_blr_tf_1))

    mean_response_1 = np.sum(tau_ax * face_on_blr_tf_1)

    assert mean_response_1 > 0
    npt.assert_almost_equal(np.sum(face_on_blr_tf_1), 1)

    # only select face on components (which will be all of them in this case)
    face_on_blr_tf_2 = calculate_blr_transfer_function(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination_angle,
        smbh_mass_exponent,
        velocity_range=[0, 1],
        weighting_grid=None,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )

    mean_response_2 = np.sum(tau_ax * face_on_blr_tf_2)

    npt.assert_almost_equal(mean_response_2, mean_response_1)

    # test that we can remove all components by selecting receeding components only
    face_on_blr_tf_3 = calculate_blr_transfer_function(
        test_density,
        vertical_vel_grid,
        radial_vel_grid,
        inclination_angle,
        smbh_mass_exponent,
        velocity_range=[-1, -0.0001],
        weighting_grid=None,
        radial_resolution=r_step,
        vertical_resolution=z_step,
    )

    mean_response_3 = np.sum(tau_ax * face_on_blr_tf_3)
    assert mean_response_3 == 0


def test_required_velocity_maximum():

    # test with easy numbers
    rest_frame_emitted_wavelength = 100
    passband_minimum = 200
    # try an absurdly high upper limit
    passband_maximum = 10**8

    redshift_0 = 0
    redshift_1 = 1

    vel_range_0 = determine_emission_line_velocities(
        rest_frame_emitted_wavelength, passband_minimum, passband_maximum, redshift_0
    )

    vel_range_1 = determine_emission_line_velocities(
        rest_frame_emitted_wavelength, passband_minimum, passband_maximum, redshift_1
    )

    assert vel_range_0[0] < vel_range_1[0]
    assert vel_range_1[1] == 0
    assert vel_range_0[1] < 0
    assert vel_range_0[0] < 0
    npt.assert_almost_equal(vel_range_0[0], vel_range_1[0])

    with pytest.raises(AssertionError):
        determine_emission_line_velocities(
            rest_frame_emitted_wavelength, passband_minimum, passband_maximum, -1
        )

    # test with required blueshift

    rest_frame_emitted_wavelength = 1000
    passband_minimum = 10 ** (-8)
    passband_maximum = 200

    vel_range_blueshift = determine_emission_line_velocities(
        rest_frame_emitted_wavelength, passband_minimum, passband_maximum, redshift_0
    )

    assert vel_range_blueshift[0] > 0
    assert vel_range_blueshift[1] > 0

    npt.assert_almost_equal(vel_range_blueshift[1], 1)

    # test with an extremely broad filter (wavelengths 10^(-8) - 10^(8))
    passband_maximum = 10**8

    vel_range_all = determine_emission_line_velocities(
        rest_frame_emitted_wavelength, passband_minimum, passband_maximum, redshift_0
    )

    npt.assert_almost_equal(vel_range_all[1], 1)
    npt.assert_almost_equal(vel_range_all[0], -1)


def test_convolve_signal_with_transfer_function():
    sample_transfer_function = [0, 0, 0, 20, 100, 80, 60, 40, 10, 0]
    smbh_mass_exp = 9.8
    driving_signal = np.sin(np.linspace(1, 1000, 1000))
    times, convolved_signal = convolve_signal_with_transfer_function(
        smbh_mass_exp=smbh_mass_exp,
        driving_signal=driving_signal,
        transfer_function=sample_transfer_function,
    )
    daily_signal = np.interp(np.linspace(1, 1000, 1000), times, convolved_signal)
    assert np.sum(abs(driving_signal - daily_signal)) != 0


def test_convert_speclite_filter_to_wavelength_range():
    # collect FilterResponse objects
    my_filters = load_filters("lsst2023-*")
    # try one FilterResponse
    one_filter = my_filters[0]
    # try one string
    one_string = "lsst2023-r"
    # try a list of strings
    string_list = ["lsst2023-g", "lsst2023-y"]

    output_one = convert_speclite_filter_to_wavelength_range(one_filter)
    output_str = convert_speclite_filter_to_wavelength_range(one_string)
    output_list = convert_speclite_filter_to_wavelength_range(string_list)

    # try an invalid string
    assert not convert_speclite_filter_to_wavelength_range("not_a_filter")
    # try something that's not a string, list, or FilterResponse
    assert not convert_speclite_filter_to_wavelength_range({"not_a_filter": 4})
    string_list_bad_names = ["one_filter", "two_filter", "red_filter", "blue"]
    assert not convert_speclite_filter_to_wavelength_range(string_list_bad_names)

    assert isinstance(output_one, list)
    assert isinstance(output_str, list)
    assert isinstance(output_list, list)
    assert isinstance(output_one[0], list)
    assert isinstance(output_str[0], list)
    assert isinstance(output_list[0], list)

    assert output_one[0][0] < output_one[0][1]
    assert output_str[0][0] < output_str[0][1]

    for item in output_list:
        assert item[0] < item[1]



import pytest
from amoeba.Classes.flux_projection import FluxProjection
from amoeba.Classes.accretion_disk import AccretionDisk
from amoeba.Classes.blr import BroadLineRegion
from amoeba.Classes.blr_streamline import Streamline
from amoeba.Util.util import create_maps
import astropy.units as u
import numpy as np
import numpy.testing as npt


def test_initialization():

    # start with completely generic maps
    # take a gaussian
    x_vals = np.linspace(-100, 100, 201)
    X, Y = np.meshgrid(x_vals, x_vals)
    R2 = X**2 + Y**2
    flux_map = np.exp(-(R2 / 20**2))

    observer_frame_wavelength_in_nm = 500
    smbh_mass_exp = 7.5
    redshift_source = 1.2
    r_out_in_gravitational_radii = 100
    inclination_angle = 0
    OmM = 0.3
    H0 = 70

    my_gaussian_projection = FluxProjection(
        flux_map,
        observer_frame_wavelength_in_nm,
        smbh_mass_exp,
        redshift_source,
        r_out_in_gravitational_radii,
        inclination_angle,
        OmM=0.3,
        H0=70,
    )

    # check the sum naturally
    natural_sum = np.sum(flux_map * my_gaussian_projection.pixel_size**2)
    assert natural_sum == my_gaussian_projection.total_flux
    assert (
        observer_frame_wavelength_in_nm
        == my_gaussian_projection.observer_frame_wavelength_in_nm
    )
    assert smbh_mass_exp == my_gaussian_projection.smbh_mass_exp
    assert redshift_source == my_gaussian_projection.redshift_source

    # test natural initialization through the AccretionDisk object

    agn_dict = create_maps(
        smbh_mass_exp,
        redshift_source,
        r_out_in_gravitational_radii,
        inclination_angle,
        100,
    )
    agn_disk = AccretionDisk(**agn_dict)

    projected_flux_object = agn_disk.calculate_surface_intensity_map(
        observer_frame_wavelength_in_nm,
    )

    assert (
        projected_flux_object.observer_frame_wavelength_in_nm
        == my_gaussian_projection.observer_frame_wavelength_in_nm
    )
    assert projected_flux_object.total_flux > 0
    assert (
        projected_flux_object.inclination_angle
        == my_gaussian_projection.inclination_angle
    )
    assert projected_flux_object.pixel_size > 0
    assert projected_flux_object.rg > 0
    assert projected_flux_object.lum_dist > 0
    assert projected_flux_object.ang_diam_dist > 0
    assert projected_flux_object.redshift_source == redshift_source

    # test construction via BLR object
    smbh_mass_exp = 7.28384
    launch_radius = 10  # Rg
    launch_theta = 0  # degrees
    max_height = 100  # Rg
    rest_frame_wavelength_in_nm = 600
    characteristic_distance = max_height // 5
    asymptotic_poloidal_velocity = 0.2
    poloidal_launch_velocity = 10**-5

    test_blr_streamline = Streamline(
        launch_radius,
        launch_theta,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        poloidal_launch_velocity=poloidal_launch_velocity,
    )

    launch_theta_angled = 45
    test_blr_streamline_angled = Streamline(
        launch_radius,
        launch_theta_angled,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        poloidal_launch_velocity=poloidal_launch_velocity,
    )

    blr = BroadLineRegion(
        smbh_mass_exp, max_height, rest_frame_wavelength_in_nm, redshift_source
    )

    blr.add_streamline_bounded_region(test_blr_streamline, test_blr_streamline_angled)

    inclination_angle = 50

    projection = blr.project_blr_total_intensity(inclination_angle)

    assert projection.flux_array.ndim == 2
    assert projection.total_flux > 0
    assert isinstance(
        projection.observer_frame_wavelength_in_nm,
        (u.Quantity, float, list, np.ndarray),
    )
    assert projection.smbh_mass_exp == smbh_mass_exp
    assert projection.r_out_in_gravitational_radii > max_height + launch_radius
    assert projection.inclination_angle == inclination_angle


def test_add_flux_projection():

    # generate two flux projections and test that we can raise errors
    x_vals = np.linspace(-100, 100, 200)
    X, Y = np.meshgrid(x_vals, x_vals)
    R2 = X**2 + Y**2
    flux_map = 200 * np.exp(-(R2 / 20**2))

    observer_frame_wavelength_in_nm = 500
    smbh_mass_exp = 7.5
    redshift_source = 1.2
    r_out_in_gravitational_radii = 100
    inclination_angle = 0
    OmM = 0.3
    H0 = 70

    my_gaussian_projection_1 = FluxProjection(
        flux_map,
        observer_frame_wavelength_in_nm,
        smbh_mass_exp,
        redshift_source,
        r_out_in_gravitational_radii,
        inclination_angle,
        OmM=0.3,
        H0=70,
    )

    wavelength_range = [23, 440]

    flux_map_2 = 500 * np.exp(-(abs(R2 - 50) / 20**2))
    my_gaussian_projection_2 = FluxProjection(
        flux_map_2,
        wavelength_range,
        smbh_mass_exp,
        redshift_source + 1,
        r_out_in_gravitational_radii,
        inclination_angle,
        OmM=0.3,
        H0=70,
    )

    assert np.sum((flux_map - flux_map_2) ** 2) > 0

    with npt.assert_raises(AssertionError):
        my_gaussian_projection_1.add_flux_projection(my_gaussian_projection_2)

    my_gaussian_projection_2.redshift_source = redshift_source
    my_gaussian_projection_2.smbh_mass_exp += 1

    with npt.assert_raises(AssertionError):
        my_gaussian_projection_1.add_flux_projection(my_gaussian_projection_2)

    my_gaussian_projection_2.smbh_mass_exp = smbh_mass_exp
    my_gaussian_projection_2.inclination_angle += 1

    with npt.assert_raises(AssertionError):
        my_gaussian_projection_1.add_flux_projection(my_gaussian_projection_2)

    my_gaussian_projection_2.inclination_angle = inclination_angle

    previous_total_flux = my_gaussian_projection_1.total_flux
    assert my_gaussian_projection_1.add_flux_projection(my_gaussian_projection_2)

    assert my_gaussian_projection_1.total_flux > previous_total_flux
    assert np.shape(my_gaussian_projection_1.flux_array) == np.shape(flux_map)

    assert np.sum((my_gaussian_projection_1.flux_array - flux_map) ** 2) > 0
    assert np.sum((my_gaussian_projection_1.flux_array - flux_map_2) ** 2) > 0

    assert (
        my_gaussian_projection_1.r_out_in_gravitational_radii
        == r_out_in_gravitational_radii
    )

    x_vals_big = np.linspace(-200, 200, 100)
    X_big, Y_big = np.meshgrid(x_vals_big, x_vals_big)
    R2_big = X_big**2 + Y_big**2
    flux_map_big = np.exp(-(R2_big / 50**2))

    observer_frame_wavelength_in_nm_2 = 800
    big_r_out_in_gravitational_radii = 200

    my_big_gaussian_projection = FluxProjection(
        flux_map_big,
        observer_frame_wavelength_in_nm_2,
        smbh_mass_exp,
        redshift_source,
        big_r_out_in_gravitational_radii,
        inclination_angle,
        OmM=0.3,
        H0=70,
    )

    low_res_rg_per_pix = (
        my_big_gaussian_projection.pixel_size / my_big_gaussian_projection.rg
    )
    high_res_rg_per_pix = (
        my_gaussian_projection_1.pixel_size / my_gaussian_projection_1.rg
    )
    small_flux_distribution_shape = np.shape(flux_map)
    big_flux_distribution_shape = np.shape(flux_map_big)

    assert low_res_rg_per_pix != high_res_rg_per_pix
    assert small_flux_distribution_shape != big_flux_distribution_shape

    assert my_gaussian_projection_1.add_flux_projection(my_big_gaussian_projection)

    output_res_rg_per_pix = (
        my_gaussian_projection_1.pixel_size / my_gaussian_projection_1.rg
    )
    output_flux_distribution_shape = np.shape(my_gaussian_projection_1.flux_array)

    assert output_res_rg_per_pix == low_res_rg_per_pix
    assert output_flux_distribution_shape == big_flux_distribution_shape

    # assure total flux is the expected value after adding projections
    assert (
        np.sum(
            my_gaussian_projection_1.flux_array * my_gaussian_projection_1.pixel_size**2
        )
        == my_gaussian_projection_1.total_flux
    )
    # assure total flux is actually the sum of all fluxes (at least to 0.01%)
    npt.assert_approx_equal(
        my_gaussian_projection_1.total_flux,
        np.sum(
            [
                previous_total_flux,
                my_gaussian_projection_2.total_flux,
                my_big_gaussian_projection.total_flux,
            ]
        ),
        4,
    )
    # make sure wavelength data is transfered as expected
    assert isinstance(my_gaussian_projection_1.observer_frame_wavelength_in_nm, list)
    min_lam = np.min(
        np.concatenate(
            (
                [observer_frame_wavelength_in_nm],
                wavelength_range,
                [observer_frame_wavelength_in_nm_2],
            )
        )
    )
    max_lam = np.max(
        np.concatenate(
            (
                [observer_frame_wavelength_in_nm],
                wavelength_range,
                [observer_frame_wavelength_in_nm_2],
            )
        )
    )
    assert min_lam == np.min(my_gaussian_projection_1.observer_frame_wavelength_in_nm)
    assert max_lam == np.max(my_gaussian_projection_1.observer_frame_wavelength_in_nm)

    assert my_big_gaussian_projection.add_flux_projection(my_gaussian_projection_1)


def test_get_plotting_axes():
    x_vals = np.linspace(-100, 100, 201)
    X, Y = np.meshgrid(x_vals, x_vals)
    R2 = X**2 + Y**2
    flux_map = np.exp(-(R2 / 20**2))

    observer_frame_wavelength_in_nm = 500
    smbh_mass_exp = 7.5
    redshift_source = 1.2
    r_out_in_gravitational_radii = 100
    inclination_angle = 0
    OmM = 0.3
    H0 = 70

    my_gaussian_projection = FluxProjection(
        flux_map,
        observer_frame_wavelength_in_nm,
        smbh_mass_exp,
        redshift_source,
        r_out_in_gravitational_radii,
        inclination_angle,
        OmM=0.3,
        H0=70,
    )

    generated_X, generated_Y = my_gaussian_projection.get_plotting_axes()

    assert np.shape(generated_X) == np.shape(generated_Y)
    assert np.shape(generated_X) == np.shape(X)
    assert np.sum(X - generated_X) == 0
    assert np.sum(Y - generated_Y) == 0



from amoeba.Util.util import convolve_signal_with_transfer_function
import numpy as np
import matplotlib.pyplot as plt
import numpy.testing as npt


def test_convolve_signal_with_transfer_function():
    tax = np.linspace(0, 1000, 1001)
    sample_tf = np.linspace(0, 1000, 1001) * np.linspace(1000, 0, 1001)
    sample_tf /= np.sum(sample_tf)

    smbh_mass_exp = 9.61

    driving_signal = np.sin(tax * np.pi / 25) * np.sin(tax * np.pi / 103)

    cadence = 0.1
    redshift_source = 1

    new_tax, new_signal = convolve_signal_with_transfer_function(
        smbh_mass_exp=smbh_mass_exp,
        driving_signal=driving_signal,
        transfer_function=sample_tf,
        initial_time_axis=tax,
        redshift_source=redshift_source,
        desired_cadence_in_days=cadence,
    )

    assert len(new_tax) == len(new_signal)
    assert len(new_tax) > len(tax)
    assert max(new_tax) == max(tax * (1 + redshift_source))

    # test that really tiny transfer functions are handled
    mass_exp = 0.1

    cadence = 1
    redshift_source = 0

    new_tax, new_signal = convolve_signal_with_transfer_function(
        smbh_mass_exp=smbh_mass_exp,
        driving_signal=driving_signal,
        transfer_function=sample_tf,
        initial_time_axis=tax,
        redshift_source=redshift_source,
        desired_cadence_in_days=cadence,
    )



import pytest
from amoeba.Classes.diffuse_continuum import DiffuseContinuum
from amoeba.Classes.flux_projection import FluxProjection
import numpy as np
import astropy.units as u
import astropy.constants as const
from amoeba.Util.util import convert_cartesian_to_polar


def test_initialization():

    inclination_angle = 30

    x_vals = np.linspace(-2000, 2000, 100)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 5
    redshift_source = 1
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 20
    r_out_in_gravitational_radii = 1000
    name = "my first diffuse continuum"

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "cloud_density_array": cloud_density_array,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
        "name": name,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    assert my_continuum.name == name
    assert my_continuum.smbh_mass_exp == smbh_mass_exp
    assert my_continuum.redshift_source == redshift_source
    assert my_continuum.inclination_angle == inclination_angle
    assert np.shape(my_continuum.radii_array) == np.shape(radii_array)
    assert (
        my_continuum.cloud_density_radial_dependence == cloud_density_radial_dependence
    )
    assert np.shape(my_continuum.cloud_density_array) == np.shape(radii_array)
    assert my_continuum.OmM == OmM
    assert my_continuum.H0 == H0
    assert my_continuum.r_in_in_gravitational_radii == r_in_in_gravitational_radii
    assert my_continuum.r_out_in_gravitational_radii == r_out_in_gravitational_radii

    my_basic_continuum = DiffuseContinuum(
        smbh_mass_exp=smbh_mass_exp,
        redshift_source=redshift_source,
        inclination_angle=inclination_angle,
        radii_array=radii_array,
        phi_array=phi_array,
        cloud_density_radial_dependence=cloud_density_radial_dependence,
        r_in_in_gravitational_radii=20,
        r_out_in_gravitational_radii=1000,
    )

    assert type(my_basic_continuum) == type(my_continuum)
    assert my_basic_continuum.smbh_mass_exp == smbh_mass_exp
    assert my_basic_continuum.redshift_source == redshift_source
    assert my_basic_continuum.inclination_angle == inclination_angle

    boosted_kwargs = my_kwargs.copy()

    lams = np.asarray([100, 120, 130, 300, 320, 700, 720, 1000])
    etas = np.asarray([0.1, 0.3, 0.15, 0.4, 0.18, 0.33, 0.22, 0.26])
    const_a = 0.3

    boosted_kwargs["emissivity_etas"] = etas
    boosted_kwargs["rest_frame_wavelengths"] = lams
    boosted_kwargs["responsivity_constant"] = const_a
    boosted_kwargs["name"] = "b00sted boi"

    my_boosted_continuum = DiffuseContinuum(**boosted_kwargs)

    assert my_boosted_continuum.name is not my_continuum.name
    assert type(my_boosted_continuum) is type(my_continuum)
    assert type(my_boosted_continuum.rest_frame_wavelengths) is not None
    assert type(my_boosted_continuum.emissivity_etas) is not None
    assert my_boosted_continuum.responsivity_constant is const_a

    with pytest.raises(ValueError):
        my_bad_continuum = DiffuseContinuum(
            smbh_mass_exp=smbh_mass_exp,
            redshift_source=redshift_source,
            inclination_angle=inclination_angle,
            radii_array=radii_array,
            phi_array=phi_array,
            cloud_density_radial_dependence=None,
            r_in_in_gravitational_radii=20,
            r_out_in_gravitational_radii=1000,
        )

    undefined_bounds_kwargs = my_kwargs.copy()
    del undefined_bounds_kwargs["r_out_in_gravitational_radii"]
    del undefined_bounds_kwargs["r_in_in_gravitational_radii"]
    my_unbound_diffuse_continuum = DiffuseContinuum(**undefined_bounds_kwargs)


def test_set_emissivity():

    inclination_angle = 30

    x_vals = np.linspace(-2000, 2000, 100)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 5
    redshift_source = 1
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 20
    r_out_in_gravitational_radii = 1000

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    lams = np.asarray([100, 120, 130, 300, 320, 700, 720, 1000])
    etas = np.asarray([0.1, 0.3, 0.15, 0.4, 0.18, 0.33, 0.22, 0.26])

    my_continuum.set_emissivity(rest_frame_wavelengths=lams, emissivity_etas=etas)

    assert np.shape(my_continuum.rest_frame_wavelengths) == np.shape(lams)
    assert np.shape(my_continuum.emissivity_etas) == np.shape(etas)


def test_set_responsivity_constant():

    inclination_angle = 30

    x_vals = np.linspace(-2000, 2000, 100)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 5
    redshift_source = 1
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 20
    r_out_in_gravitational_radii = 1000

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    const_a = 0.2

    my_continuum.set_responsivity_constant(responsivity_constant=const_a)

    assert my_continuum.responsivity_constant == const_a

    with pytest.raises(AssertionError):
        my_continuum.set_responsivity_constant(responsivity_constant=-1)
    with pytest.raises(AssertionError):
        my_continuum.set_responsivity_constant(responsivity_constant=33)


def test_interpolate_spectrum_to_wavelength():

    inclination_angle = 30

    x_vals = np.linspace(-2000, 2000, 100)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 5
    redshift_source = 1
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 20
    r_out_in_gravitational_radii = 1000

    lams = [0, 100]
    etas = [0, 1]

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "cloud_density_array": cloud_density_array,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
        "emissivity_etas": etas,
        "rest_frame_wavelengths": lams,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    zero_percentile = 0 * (1 + redshift_source)
    fifty_percentile = 50 * (1 + redshift_source)
    hundred_percentile = 100 * (1 + redshift_source)

    val_0 = my_continuum.interpolate_spectrum_to_wavelength(
        zero_percentile,
    )

    val_50 = my_continuum.interpolate_spectrum_to_wavelength(fifty_percentile)

    val_100 = my_continuum.interpolate_spectrum_to_wavelength(hundred_percentile)

    assert val_0 == 0
    assert val_50 == 0.5
    assert val_100 == 1.0

    lams = [0, 100]
    etas = [2, 5]

    my_continuum.set_emissivity(rest_frame_wavelengths=lams, emissivity_etas=etas)

    assert val_0 != my_continuum.interpolate_spectrum_to_wavelength(
        zero_percentile * (1 + redshift_source)
    )

    assert my_continuum.interpolate_spectrum_to_wavelength(
        100 * (1 + redshift_source)
    ) == my_continuum.interpolate_spectrum_to_wavelength(150 * (1 + redshift_source))

    assert my_continuum.interpolate_spectrum_to_wavelength(
        0
    ) == my_continuum.interpolate_spectrum_to_wavelength(-20)


def test_get_diffuse_continuum_emission():

    inclination_angle = 30

    x_vals = np.linspace(-2000, 2000, 100)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 5
    redshift_source = 1
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 20
    r_out_in_gravitational_radii = 1000

    lams = [0, 100]
    etas = [0, 1]

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "cloud_density_array": cloud_density_array,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
        "emissivity_etas": etas,
        "rest_frame_wavelengths": lams,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    wavelength_obs = 40

    my_projection = my_continuum.get_diffuse_continuum_emission(wavelength_obs)

    assert isinstance(my_projection, FluxProjection)

    assert my_projection.redshift_source == redshift_source
    assert my_projection.rest_frame_wavelength_in_nm == wavelength_obs / (
        1 + redshift_source
    )
    assert my_projection.smbh_mass_exp == smbh_mass_exp
    assert my_projection.mass == my_continuum.mass
    assert my_projection.r_out_in_gravitational_radii == r_out_in_gravitational_radii
    assert my_projection.inclination_angle == inclination_angle
    assert my_projection.rg == my_continuum.rg
    assert my_projection.OmM == OmM
    assert my_projection.H0 == H0
    assert my_projection.lum_dist == my_continuum.lum_dist


def test_get_diffuse_continuum_mean_lag():

    inclination_angle = 0

    x_vals = np.linspace(-100, 100, 1000)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 8
    redshift_source = 2
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 30
    r_out_in_gravitational_radii = 100

    first_mask = R >= r_in_in_gravitational_radii
    second_mask = R <= r_out_in_gravitational_radii
    r_mask = np.logical_and(first_mask, second_mask)
    R *= r_mask

    lams = [100, 300, 1000]
    etas = [0.2, 0.45, 0.3]

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "cloud_density_array": cloud_density_array,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
        "emissivity_etas": etas,
        "rest_frame_wavelengths": lams,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    wavelength_obs = 400

    mean_lag = my_continuum.get_diffuse_continuum_mean_lag(wavelength_obs)

    assert isinstance(mean_lag, float)
    assert mean_lag > 0

    my_continuum.cloud_density_radial_dependence = -1

    new_mean_lag = my_continuum.get_diffuse_continuum_mean_lag(wavelength_obs)

    assert isinstance(new_mean_lag, float)
    assert new_mean_lag > 0
    assert new_mean_lag < mean_lag

    my_continuum.cloud_density_radial_dependence = 1

    large_mean_lag = my_continuum.get_diffuse_continuum_mean_lag(wavelength_obs)

    assert isinstance(large_mean_lag, float)
    assert large_mean_lag > mean_lag

    # test the direct method of numerical integration
    cloud_density_radial_dependence = None
    cloud_density_array = np.ones(np.shape(R))
    my_kwargs["cloud_density_radial_dependence"] = cloud_density_radial_dependence
    my_kwargs["cloud_density_array"] = cloud_density_array
    my_kwargs["r_out_in_gravitational_radii"] = 50

    brute_continuum = DiffuseContinuum(**my_kwargs)

    brute_force_continuum_lag = brute_continuum.get_diffuse_continuum_mean_lag(
        wavelength_obs
    )

    assert isinstance(brute_force_continuum_lag, float)
    # assure better than 5% convergence for numerical integration
    percent_diff = abs(brute_force_continuum_lag - mean_lag) / mean_lag
    assert percent_diff <= 0.05


def test_get_diffuse_continuum_lag_contribution():

    inclination_angle = 0

    x_vals = np.linspace(-100, 100, 1000)
    X, Y = np.meshgrid(x_vals, x_vals / np.cos(inclination_angle * np.pi / 180))
    R, Phi = convert_cartesian_to_polar(X, Y)

    smbh_mass_exp = 8
    redshift_source = 2
    radii_array = R
    phi_array = Phi
    cloud_density_radial_dependence = 0
    cloud_density_array = None
    OmM = 0.4
    H0 = 80
    r_in_in_gravitational_radii = 30
    r_out_in_gravitational_radii = 100

    first_mask = R >= r_in_in_gravitational_radii
    second_mask = R <= r_out_in_gravitational_radii
    r_mask = np.logical_and(first_mask, second_mask)
    R *= r_mask

    lams = [100, 300, 1000]
    etas = [0.2, 0.45, 0.3]

    constant_a = 0

    my_kwargs = {
        "smbh_mass_exp": smbh_mass_exp,
        "redshift_source": redshift_source,
        "inclination_angle": inclination_angle,
        "radii_array": radii_array,
        "phi_array": phi_array,
        "cloud_density_radial_dependence": cloud_density_radial_dependence,
        "cloud_density_array": cloud_density_array,
        "OmM": OmM,
        "H0": H0,
        "r_in_in_gravitational_radii": r_in_in_gravitational_radii,
        "r_out_in_gravitational_radii": r_out_in_gravitational_radii,
        "emissivity_etas": etas,
        "rest_frame_wavelengths": lams,
        "responsivity_constant": constant_a,
    }

    my_continuum = DiffuseContinuum(**my_kwargs)

    wavelength_obs = 400

    diffuse_continuum_additional_lag = (
        my_continuum.get_diffuse_continuum_lag_contribution(wavelength_obs)
    )

    diffuse_tau = my_continuum.get_diffuse_continuum_mean_lag(wavelength_obs)

    assert isinstance(diffuse_continuum_additional_lag, float)
    assert diffuse_continuum_additional_lag > 0
    assert diffuse_continuum_additional_lag / diffuse_tau < 1

    # delay_peak_wavelength = lams[argmax[etas]]. These are rest frame.
    longest_delay_wavelength = lams[np.argmax(np.asarray(etas))]
    minimum_delay_wavelength = lams[np.argmin(np.asarray(etas))]
    middle_wavelength = (
        minimum_delay_wavelength
        + (longest_delay_wavelength - minimum_delay_wavelength) / 2
    )
    quarter_wavelenth = (
        minimum_delay_wavelength + (middle_wavelength - minimum_delay_wavelength) / 2
    )

    longest_delay_wavelength *= 1 + redshift_source
    minimum_delay_wavelength *= 1 + redshift_source
    middle_wavelength *= 1 + redshift_source
    quarter_wavelenth *= 1 + redshift_source

    longest_delay = my_continuum.get_diffuse_continuum_lag_contribution(
        longest_delay_wavelength
    )
    minimum_delay = my_continuum.get_diffuse_continuum_lag_contribution(
        minimum_delay_wavelength
    )
    middle_delay = my_continuum.get_diffuse_continuum_lag_contribution(
        middle_wavelength
    )
    quarter_delay = my_continuum.get_diffuse_continuum_lag_contribution(
        quarter_wavelenth
    )

    assert minimum_delay < longest_delay
    assert minimum_delay < middle_delay
    assert minimum_delay < quarter_delay
    assert middle_delay < longest_delay
    assert quarter_delay < middle_delay



from astropy import units as u
from astropy import constants as const
import numpy as np
from amoeba.Util.util import (
    calculate_gravitational_radius,
    calculate_luminosity_distance,
    calculate_angular_diameter_distance,
)
from skimage.transform import rescale
from math import isinf


class FluxProjection:

    def __init__(
        self,
        flux_array,
        observer_frame_wavelength_in_nm,
        smbh_mass_exp,
        redshift_source,
        r_out_in_gravitational_radii,
        inclination_angle,
        OmM=0.3,
        H0=70,
        **kwargs
    ):
        """Initialize the projection. This object carries the projected array as well as
        important metadata.

        :param flux_array: a 2 dimensional representation of the flux array in arbitrary
            units. Spacing along the X and Y direction must be even.
        :param observer_frame_wavelength_in_nm: Wavelength or range of wavelengths this
            flux projection is associated with, in units nm, in the observer's frame of
            reference.
        :param smbh_mass_exp: solution to log10(m_smbh / m_sun), determines the distance
            of one gravitational radius and therefore sets the size scale of the
            flux_array
        :param redshift_source: redshift of the object
        :param r_out_in_gravitational_radii: maximum radius of the flux_array along each
            axis in units of R_g.
        :param inclination_angle: inclination of the source in degrees
        :param OmM: mass contribution to the energy budget of the universe
        :param H0: Hubble constant in units km/s/Mpc
        """

        self.flux_array = np.asarray(flux_array)
        self.observer_frame_wavelength_in_nm = observer_frame_wavelength_in_nm
        if isinstance(observer_frame_wavelength_in_nm, (list, np.ndarray)):
            min_wavelength = round(
                observer_frame_wavelength_in_nm[0] / (1 + redshift_source)
            )
            if not isinf(observer_frame_wavelength_in_nm[1]):
                max_wavelength = round(
                    observer_frame_wavelength_in_nm[1] / (1 + redshift_source)
                )
            else:
                max_wavelength = observer_frame_wavelength_in_nm[1]
            rest_frame_wavelength_in_nm = [
                min_wavelength,
                max_wavelength,
            ]
        else:
            rest_frame_wavelength_in_nm = round(
                observer_frame_wavelength_in_nm / (1 + redshift_source)
            )
        self.rest_frame_wavelength_in_nm = rest_frame_wavelength_in_nm
        self.smbh_mass_exp = smbh_mass_exp
        self.mass = 10**smbh_mass_exp * const.M_sun.to(u.kg)
        self.r_out_in_gravitational_radii = r_out_in_gravitational_radii
        self.redshift_source = redshift_source
        self.inclination_angle = inclination_angle
        self.rg = calculate_gravitational_radius(10**self.smbh_mass_exp)
        self.pixel_size = (
            2 * self.rg * self.r_out_in_gravitational_radii / np.size(flux_array, 0)
        )
        self.total_flux = np.sum(self.flux_array * self.pixel_size**2)
        self.OmM = OmM
        self.H0 = H0
        self.lum_dist = calculate_luminosity_distance(
            self.redshift_source, OmM=self.OmM, H0=self.H0
        )
        self.ang_diam_dist = calculate_angular_diameter_distance(
            self.redshift_source, OmM=self.OmM, H0=self.H0
        )

    def add_flux_projection(self, SecondProjection):
        """Join this FluxProjection with another FluxProjection via simple addition. The
        resulting FluxProjection will have pixels the size of the larger FluxProjection
        and a maximum radius of the larger FluxProjection. Note that these are not
        necessarily the same (a smaller and coursely spaced projection joined with a
        large and finely spaced projection will result in a large and coursely spaced
        projection).

        :param SecondProjection: FluxProjection object to add to the first projection
        :return: True if successful. Note that this method modifies the parent
            FluxProjection.
        """
        assert self.redshift_source == SecondProjection.redshift_source
        assert self.smbh_mass_exp == SecondProjection.smbh_mass_exp
        assert self.inclination_angle == SecondProjection.inclination_angle

        desired_pixel_size_in_rg = np.max(
            [self.pixel_size / self.rg, SecondProjection.pixel_size / self.rg]
        )

        working_flux_projection = SecondProjection.flux_array.copy()

        self_resolution_ratio = (self.pixel_size / self.rg) / desired_pixel_size_in_rg
        second_resolution_ratio = (
            SecondProjection.pixel_size / self.rg
        ) / desired_pixel_size_in_rg

        self.flux_array = rescale(self.flux_array, self_resolution_ratio)
        self.pixel_size = desired_pixel_size_in_rg * self.rg

        working_flux_projection = rescale(
            working_flux_projection, second_resolution_ratio
        )

        new_max_radius = np.max(
            [
                self.r_out_in_gravitational_radii,
                SecondProjection.r_out_in_gravitational_radii,
            ]
        )

        desired_flux_array_shape = (
            int(2 * new_max_radius / desired_pixel_size_in_rg),
            int(2 * new_max_radius / desired_pixel_size_in_rg),
        )

        if self.r_out_in_gravitational_radii != new_max_radius:
            increase_in_pixels = int(
                (new_max_radius - self.r_out_in_gravitational_radii)
                // desired_pixel_size_in_rg
            )
            self.flux_array = np.pad(self.flux_array, increase_in_pixels)
        if SecondProjection.r_out_in_gravitational_radii != new_max_radius:
            increase_in_pixels = int(
                (new_max_radius - SecondProjection.r_out_in_gravitational_radii)
                // desired_pixel_size_in_rg
            )
            working_flux_projection = np.pad(
                working_flux_projection, increase_in_pixels
            )

        if np.shape(self.flux_array) != np.shape(working_flux_projection):
            x_diff = np.size(self.flux_array, 0) - np.size(working_flux_projection, 0)
            y_diff = np.size(self.flux_array, 1) - np.size(working_flux_projection, 1)

            if x_diff < 0:  # pragma: no cover
                working_flux_projection = working_flux_projection[:x_diff, :]
            elif x_diff > 0:  # pragma: no cover
                working_flux_projection = np.pad(
                    working_flux_projection, ((0, x_diff), (0, 0))
                )
            if y_diff < 0:  # pragma: no cover
                working_flux_projection = working_flux_projection[:, :y_diff]
            elif y_diff > 0:  # pragma: no cover
                working_flux_projection = np.pad(
                    working_flux_projection, ((0, 0), (0, y_diff))
                )

        self.r_out_in_gravitational_radii = new_max_radius
        self.flux_array += working_flux_projection
        self.total_flux = np.sum(self.flux_array * self.pixel_size**2)

        if not isinstance(self.observer_frame_wavelength_in_nm, list):
            lams_1 = [self.observer_frame_wavelength_in_nm]
        else:
            lams_1 = self.observer_frame_wavelength_in_nm
        if not isinstance(SecondProjection.observer_frame_wavelength_in_nm, list):
            lams_2 = [SecondProjection.observer_frame_wavelength_in_nm]
        else:
            lams_2 = SecondProjection.observer_frame_wavelength_in_nm

        concatenated_lams = np.concatenate((lams_1, lams_2))
        self.observer_frame_wavelength_in_nm = [
            np.min(concatenated_lams),
            np.max(concatenated_lams),
        ]

        return True

    def get_plotting_axes(self):
        """Convenience method to get the X and Y axes to plot the flux_projection in
        physical units.

        :return: X and Y arrays in units of R_g used to create contour plots via
            plt.contourf(X, Y, self.flux_array).
        """
        xax = np.linspace(
            -self.r_out_in_gravitational_radii,
            self.r_out_in_gravitational_radii,
            np.size(self.flux_array, 0),
        )
        X, Y = np.meshgrid(xax, xax)
        return X, Y



from astropy import units as u
from astropy import constants as const
import numpy as np
from astropy.io import fits
from amoeba.Util.util import (
    calculate_angular_diameter_distance,
    calculate_einstein_radius_in_meters,
    convert_1d_array_to_2d_array,
    perform_microlensing_convolution,
    pull_value_from_grid,
    extract_light_curve,
    calculate_microlensed_transfer_function,
)


class MagnificationMap:

    def __init__(
        self,
        redshift_source,
        redshift_lens,
        magnification_array,
        convergence,
        shear,
        mean_microlens_mass_in_kg=1 * const.M_sun.to(u.kg),
        total_microlens_einstein_radii=25,
        OmM=0.3,
        H0=70,
        name="",
        **kwargs
    ):
        """This class is for microlensing magnification maps, which should be 2d
        representations of the magnification at each pixel. The magnification map can be
        input as either magnification_array or ray counts.

        :param redshift_source: Redshift of the distant source being lensed
        :param redshift_lens: Redshift of the lensing galaxy
        :param magnification_array: 2d numpy array, fits file, or binary file
            representing the magnification map
        :param convergence: Convergence of the lensing potential at the location of the
            image
        :param shear: Shear of the lensing potential at the location of the image
        :param mean_microlens_mass_in_kg: Average microlens mass in kg
        :param total_microlens_einstein_radii: Number of Einstein radii the
            magnification map covers along one square edge.
        :param OmM: Cosmological fraction of mass in the energy budget of the universe
        :param H0: Hubble constant in units km/s/Mpc
        :param name: Name space.
        """

        self.name = name
        self.redshift_source = redshift_source
        self.redshift_lens = redshift_lens
        self.convergence = convergence
        self.shear = shear
        self.total_microlens_einstein_radii = total_microlens_einstein_radii
        self.mean_microlens_mass_in_kg = mean_microlens_mass_in_kg
        self.OmM = OmM
        self.H0 = H0
        self.einstein_radius_in_meters = calculate_einstein_radius_in_meters(
            self.redshift_lens,
            self.redshift_source,
            mean_microlens_mass_in_kg=self.mean_microlens_mass_in_kg,
            OmM=OmM,
            H0=self.H0,
        )

        if isinstance(magnification_array, (np.ndarray, list)):
            if np.ndim(magnification_array) == 1:
                self.ray_map = convert_1d_array_to_2d_array(magnification_array)
            elif np.ndim(magnification_array) == 2:
                self.ray_map = magnification_array
        elif magnification_array[-4:] == "fits":  # pragma: no cover
            with fits.open(magnification_array) as f:
                self.ray_map = f[0].data
        elif (
            magnification_array[-4:] == ".dat" or magnification_array[-4:] == ".bin"
        ):  # pragma: no cover
            with open(magnification_array, "rb") as f:
                extracted_magnification_array = np.fromfile(f, "i", count=-1, sep="")
                self.ray_map = convert_1d_array_to_2d_array(
                    extracted_magnification_array
                )
        else:
            print(
                "Invalid file name / format. Please pass in the pathway to a .fits or .dat file or a Numpy array"
            )
            return None

        self.resolution = np.size(self.ray_map, 0)
        self.macro_magnification = 1 / ((1 - self.convergence) ** 2.0 - self.shear**2.0)

        self.ray_to_mag_ratio = np.sum(self.ray_map) / np.size(self.ray_map)
        self.magnification_array = self.ray_map / self.ray_to_mag_ratio

        self.pixel_size = (
            self.einstein_radius_in_meters
            * self.total_microlens_einstein_radii
            / self.resolution
        )
        self.pixel_shift = 0

    def convolve_with_flux_projection(
        self, FluxProjection, relative_orientation=False, random_seed=None
    ):
        """Prepare the convolution between this magnification map and a FluxProjection.
        Note that once the convolution is performed, all orientations are defined. In
        particular, this solidifies the orientation of the source w.r.t. the direction
        of shear of the magnification map.

        :param FluxProjection: FluxProjection object representing the source plane flux
            density.
        :param relative_orientation: Angle to rotate the projection with respect to the
            magnification map in degrees.
        :param random_seed: A random seed to set random parameters consistently
        :return: Child magnification map object representing all magnifications of the
            flux projection. Normalization of coordinates w.r.t. the smbh is considered,
            and you will probably use this primarily for its "pull_light_curve" method.
        """

        convolved_map = ConvolvedMap(
            self, FluxProjection, relative_orientation=False, random_seed=random_seed
        )

        return convolved_map

    def pull_value_from_grid(self, x_val, y_val, weight_by_macromag=False):
        """Read off the magnification at some position.

        :param x_val: X value on the magnification map in pixels
        :param y_val: Y value on the magnification map in pixels
        :param weight_by_macromag: boolean toggle to weight the values by the
            magnification map's macro magnification. This macro magnification is
            determined by the convergence and shear.
        :return: Magnification at location (x_val, y_val).
        """

        value = pull_value_from_grid(
            self.magnification_array, x_val + self.pixel_shift, y_val + self.pixel_shift
        )

        if weight_by_macromag:
            value *= self.macro_magnification
        return value

    def pull_light_curve(
        self,
        effective_transverse_velocity,
        light_curve_duration_in_years,
        x_start_position=None,
        y_start_position=None,
        phi_travel_direction=None,
        weight_by_macromag=False,
        return_track_coords=False,
        random_seed=None,
    ):
        """Calculate the values of a light curve. By default, the light curve will be
        randomly generated from the region of the magnification map not affected by
        convolutional artifacts.

        :param effective_transverse_velocity: effective transverse velocity in km/s. This
            is typically on the order of ~10^2 - 10^3 km/s. However, exceptional cases
            may exceeed 10^4 km/s.
        :param light_curve_duration_in_years: duration of light curve in years
        :param x_start_position: x coordinate to start at in pixels
        :param y_start_position: y coordinate to start at in pixels
        :param phi_travel_direction: direction to travel in, in degrees
        :param weight_by_macromag: boolean toggle to weight the values by the magnification
            map's macro magnification. This macro magnification is determined by the
            convergence and shear.
        :param return_track_coords: Bool to return the pixel locations as well as the
            light curve
        :param random_seed: Set a random seed so random values are determined
            consistently

        :return: Either a list representing the light curve or (for return_track_coords) a
            tuple representing a list for the light curve, and two lists for the x, y positions
        """

        light_curve = extract_light_curve(
            self.magnification_array,
            self.pixel_size,
            effective_transverse_velocity,
            light_curve_duration_in_years,
            pixel_shift=self.pixel_shift,
            x_start_position=x_start_position,
            y_start_position=y_start_position,
            phi_travel_direction=phi_travel_direction,
            return_track_coords=return_track_coords,
            random_seed=random_seed,
        )
        if weight_by_macromag:
            if not return_track_coords:
                light_curve *= self.macro_magnification
            else:
                track_x = light_curve[1]
                track_y = light_curve[2]
                light_curve = self.macro_magnification * np.asarray(light_curve[0])
                light_curve = light_curve, track_x, track_y
        return light_curve

    def calculate_microlensed_transfer_function(
        self,
        AccretionDisk,
        observer_frame_wavelength_in_nm,
        corona_height=None,
        relative_orientation=0,
        x_position=None,
        y_position=None,
        axis_offset_in_gravitational_radii=0,
        angle_offset_in_degrees=0,
        OmM=0.3,
        H0=70,
        return_response_array_and_lags=False,
        return_descaled_response_array_and_lags=False,
        return_magnification_map_crop=False,
        random_seed=None,
    ):
        """Calculate the transfer function of an accretion disk when the response map is
        microlensed at a particular location.

        :param AccretionDisk: AccretionDisk object
        :param observer_frame_wavelength_in_nm: observer frame wavelength in nanometers
            to calculate the response at
        :param corona_height: None or int/float. If none, will take the value stored in
            the AccretionDisk object. Otherwise will override the coronaheight in units
            R_g = GM/c^2
        :param relative_orientation: Degree rotation of the AccretionDisk with respect
            to the magnification map
        :param x_position: None or pixel position of the disk. If None, a random
            position is used
        :param y_position: None or pixel position of the disk. If None, a random
            position is used
        :param axis_offset_in_gravitational_radii: Off axis position to calculate the
            response map from
        :param angle_offset_in_degrees: Degree rotation of the axis offset
        :param OmM: mass component of the energy budget of the universe
        :param H0: Hubble constant in units km/s/Mpc
        :param return_response_array_and_lags: Boolean used to return the spatially
            resolved maps associated with the response function.
        :param return_descaled_response_array_and_lags: Similar to above, but returns
            the projections at the resolution of the magnification map. Useful for
            debugging.
        :param return_magnification_map_crop: Boolean used to return the region of
            the magnification map used to amplify the response function.
        :param random_seed: allows the user to set a random seed for reproducibility
        :return: The microlensed transfer function represented by a list.
        """

        if corona_height is None:
            corona_height = AccretionDisk.corona_height

        rest_wavelength_in_nm = observer_frame_wavelength_in_nm / (
            1 + self.redshift_source
        )

        return calculate_microlensed_transfer_function(
            self.magnification_array,
            self.redshift_lens,
            AccretionDisk.redshift_source,
            rest_wavelength_in_nm,
            AccretionDisk.temp_array,
            AccretionDisk.radii_array,
            AccretionDisk.phi_array,
            AccretionDisk.g_array,
            AccretionDisk.inclination_angle,
            AccretionDisk.smbh_mass_exp,
            corona_height,
            mean_microlens_mass_in_kg=self.mean_microlens_mass_in_kg,
            number_of_microlens_einstein_radii=self.total_microlens_einstein_radii,
            number_of_smbh_gravitational_radii=AccretionDisk.r_out_in_gravitational_radii,
            relative_orientation=relative_orientation,
            OmM=OmM,
            H0=H0,
            axis_offset_in_gravitational_radii=axis_offset_in_gravitational_radii,
            angle_offset_in_degrees=angle_offset_in_degrees,
            height_array=AccretionDisk.height_array,
            albedo_array=AccretionDisk.albedo_array,
            x_position=x_position,
            y_position=y_position,
            return_response_array_and_lags=return_response_array_and_lags,
            return_descaled_response_array_and_lags=return_descaled_response_array_and_lags,
            return_magnification_map_crop=return_magnification_map_crop,
            random_seed=random_seed,
        )

class ConvolvedMap(MagnificationMap):
    """This child class represents the convoultion of a magnification map with a surface
    flux density.

    Has all of the same methods as MagnificationMap
    """

    def __init__(
        self,
        magnification_map,
        projected_flux_distribution,
        relative_orientation=False,
        random_seed=None,
    ):

        output_convolution, pixel_shift = perform_microlensing_convolution(
            magnification_map.magnification_array,
            projected_flux_distribution.flux_array,
            redshift_lens=magnification_map.redshift_lens,
            redshift_source=magnification_map.redshift_source,
            smbh_mass_exp=projected_flux_distribution.smbh_mass_exp,
            mean_microlens_mass_in_kg=magnification_map.mean_microlens_mass_in_kg,
            number_of_microlens_einstein_radii=magnification_map.total_microlens_einstein_radii,
            number_of_smbh_gravitational_radii=projected_flux_distribution.r_out_in_gravitational_radii,
            relative_orientation=relative_orientation,
            OmM=magnification_map.OmM,
            H0=magnification_map.H0,
            random_seed=random_seed,
        )

        self.pixel_size = magnification_map.pixel_size
        self.total_microlens_einstein_radii = (
            magnification_map.total_microlens_einstein_radii
        )
        self.mean_microlens_mass_in_kg = magnification_map.mean_microlens_mass_in_kg
        self.resolution = magnification_map.resolution
        self.magnification_array = output_convolution
        self.pixel_shift = pixel_shift
        self.macro_magnification = magnification_map.macro_magnification
        self.redshift_lens = magnification_map.redshift_lens

        self.smbh_mass_exp = projected_flux_distribution.smbh_mass_exp
        self.inclination_angle = projected_flux_distribution.inclination_angle
        self.rg = projected_flux_distribution.rg
        self.observer_frame_wavelength_in_nm = (
            projected_flux_distribution.observer_frame_wavelength_in_nm
        )
        self.relative_orientation = relative_orientation
        self.redshift_source = projected_flux_distribution.redshift_source



import numpy as np
import numpy.testing as npt
from amoeba.Classes.magnification_map import MagnificationMap, ConvolvedMap
from amoeba.Classes.flux_projection import FluxProjection
from amoeba.Classes.accretion_disk import AccretionDisk
from amoeba.Util.util import create_maps
import pytest


class TestMagnificationMap:
    def setup_method(self):
        magnification_array = [
            [1, 2, 1, 1, 1, 2, 3, 4, 3],
            [1, 3, 2, 1, 1, 2, 3, 4, 2],
            [1, 5, 8, 2, 1, 2, 3, 4, 1],
            [1, 6, 16, 8, 2, 2, 3, 4, 2],
            [1, 4, 8, 3, 1, 2, 3, 4, 3],
            [1, 3, 1, 1, 1, 2, 3, 5, 3],
            [4, 2, 6, 3, 3, 2, 5, 10, 5],
            [5, 2, 7, 20, 7, 2, 3, 5, 3],
            [4, 2, 12, 15, 12, 2, 3, 4, 3],
        ]
        flat_array = np.asarray(magnification_array.copy()).flatten()
        assert np.ndim(flat_array) == 1

        redshift_source = 2.0
        redshift_lens = 1.0
        convergence = 0.3
        shear = 0.1
        name = "silly test array"
        total_microlens_einstein_radii = 1

        self.magnification_map = MagnificationMap(
            redshift_source,
            redshift_lens,
            magnification_array,
            convergence,
            shear,
            total_microlens_einstein_radii=total_microlens_einstein_radii,
            name=name,
        )

        alt_magnification_map = MagnificationMap(
            redshift_source,
            redshift_lens,
            flat_array,
            convergence,
            shear,
            total_microlens_einstein_radii=total_microlens_einstein_radii,
            name=name,
        )
        # Note I am not including tests for the lines which open magnification
        # map files. To do so, I would have to include some test files or
        # design the test function to first write these files then delete them.

        MagnificationMap(
            redshift_source,
            redshift_lens,
            "definitely_not_a_file_path.invalid_type",
            convergence,
            shear,
            total_microlens_einstein_radii=total_microlens_einstein_radii,
            name=name,
        )

        assert self.magnification_map.name == name
        assert self.magnification_map.redshift_source == redshift_source
        assert self.magnification_map.redshift_lens == redshift_lens
        assert self.magnification_map.shear == shear
        assert self.magnification_map.convergence == convergence
        assert self.magnification_map.H0 == 70
        assert self.magnification_map.einstein_radius_in_meters > 0
        assert self.magnification_map.resolution == 9
        assert self.magnification_map.macro_magnification == (
            1 / ((1 - convergence) ** 2 - shear**2)
        )
        assert self.magnification_map.pixel_size > 0
        # demand that it was normalized
        assert np.sum(self.magnification_map.magnification_array) == np.size(
            self.magnification_map.magnification_array
        )
        assert np.argmax(self.magnification_map.ray_map) == np.argmax(
            self.magnification_map.magnification_array
        )

    def test_pull_value_from_grid(self):
        x_value = 3
        y_value = 5

        magnification_value = self.magnification_map.pull_value_from_grid(
            x_value, y_value
        )

        assert magnification_value > 0

        x_value = 1.1
        y_value = 6.8

        magnification_value_decimal = self.magnification_map.pull_value_from_grid(
            x_value, y_value
        )

        assert magnification_value_decimal > 0

        magnification_value_decimal_macromag = (
            self.magnification_map.pull_value_from_grid(
                x_value, y_value, weight_by_macromag=True
            )
        )

        assert magnification_value_decimal_macromag > 0

        assert magnification_value_decimal_macromag != magnification_value_decimal

    def test_pull_light_curve(self):

        stationary_transverse_velocity = 0.0
        effective_transverse_velocity = 4000
        light_curve_duration_in_years = 1

        x_start_position = 5
        y_start_position = 5

        random_seed = 4

        stationary_light_curve = self.magnification_map.pull_light_curve(
            stationary_transverse_velocity,
            light_curve_duration_in_years,
            x_start_position=x_start_position,
            y_start_position=y_start_position,
            random_seed=random_seed,
        )

        light_curve = self.magnification_map.pull_light_curve(
            effective_transverse_velocity,
            light_curve_duration_in_years,
            x_start_position=x_start_position,
            y_start_position=y_start_position,
            random_seed=random_seed,
        )

        # a light curve must be at least 2 points
        assert len(light_curve) >= 2
        # light curve should evolve, while stationary will not
        assert stationary_light_curve[0] == stationary_light_curve[-1]
        assert stationary_light_curve[-1] != light_curve[-1]
        # they have the same start point
        assert light_curve[0] == stationary_light_curve[0]

        # define a path which goes over a high magnification point
        y_start_position = 4
        x_start_position = 1
        phi_travel_direction = 90

        light_curve_rise_fall = self.magnification_map.pull_light_curve(
            effective_transverse_velocity,
            light_curve_duration_in_years,
            x_start_position=x_start_position,
            y_start_position=y_start_position,
            phi_travel_direction=phi_travel_direction,
        )

        assert len(light_curve_rise_fall) >= 4
        assert light_curve_rise_fall[2] > light_curve_rise_fall[0]

        light_curve_macromag = self.magnification_map.pull_light_curve(
            effective_transverse_velocity,
            light_curve_duration_in_years,
            x_start_position=x_start_position,
            y_start_position=y_start_position,
            random_seed=random_seed,
            weight_by_macromag=True,
        )
        light_curve_macromag_components = self.magnification_map.pull_light_curve(
            effective_transverse_velocity,
            light_curve_duration_in_years,
            x_start_position=x_start_position,
            y_start_position=y_start_position,
            random_seed=random_seed,
            return_track_coords=True,
            weight_by_macromag=True,
        )

        assert (
            np.sum(abs(light_curve_macromag - light_curve_macromag_components[0])) == 0
        )

    def test_convolve_with_flux_projection(self):

        # must generate a simple flux projection object to test this method
        flux_array = [[1 / 4, 1 / 2, 1 / 4], [1 / 2, 1, 1 / 2], [1 / 4, 1 / 2, 1 / 4]]
        observer_wavelength = 100
        smbh_mass_exp = 6.9
        redshift_source = self.magnification_map.redshift_source
        r_out_in_rg = 700
        inclination = 0

        test_flux_projection = FluxProjection(
            flux_array,
            observer_wavelength,
            smbh_mass_exp,
            redshift_source,
            r_out_in_rg,
            inclination,
        )

        self.test_convolution_object = (
            self.magnification_map.convolve_with_flux_projection(test_flux_projection)
        )

        # check inheritance
        assert isinstance(self.test_convolution_object, MagnificationMap)
        assert isinstance(self.test_convolution_object, ConvolvedMap)

        # check everything passed through properly
        assert (
            self.magnification_map.redshift_source
            == self.test_convolution_object.redshift_source
        )
        assert (
            self.magnification_map.pixel_size == self.test_convolution_object.pixel_size
        )
        assert (
            self.magnification_map.total_microlens_einstein_radii
            == self.test_convolution_object.total_microlens_einstein_radii
        )
        assert (
            self.magnification_map.mean_microlens_mass_in_kg
            == self.test_convolution_object.mean_microlens_mass_in_kg
        )
        assert (
            self.magnification_map.resolution == self.test_convolution_object.resolution
        )
        assert (
            test_flux_projection.smbh_mass_exp
            == self.test_convolution_object.smbh_mass_exp
        )
        assert (
            test_flux_projection.inclination_angle
            == self.test_convolution_object.inclination_angle
        )
        assert (
            test_flux_projection.observer_frame_wavelength_in_nm
            == self.test_convolution_object.observer_frame_wavelength_in_nm
        )
        assert (
            self.magnification_map.redshift_lens
            == self.test_convolution_object.redshift_lens
        )
        assert (
            self.magnification_map.macro_magnification
            == self.test_convolution_object.macro_magnification
        )

        # check the convolution actually changed the values
        x_value = 3
        y_value = np.pi

        assert self.test_convolution_object.pull_value_from_grid(
            x_value,
            y_value,
        ) != self.magnification_map.pull_value_from_grid(
            x_value,
            y_value,
        )

    def test_calculate_microlensed_transfer_function(self):

        # need to set up an accretion disk object
        smbh_mass_exp = 8.0
        redshift_source = 2.0
        inclination_angle = 0.0
        corona_height = 10
        number_grav_radii = 500
        resolution = 500
        spin = 0

        accretion_disk_data_1 = create_maps(
            smbh_mass_exp,
            redshift_source,
            number_grav_radii,
            inclination_angle,
            resolution,
            spin=spin,
            corona_height=corona_height,
        )

        self.AccretionDisk = AccretionDisk(**accretion_disk_data_1)

        # as a sanity check, prepare an identity array for microlensing

        big_magnification_ones = np.ones((resolution, resolution))

        self.identity_magnification_array = MagnificationMap(
            redshift_source,
            self.magnification_map.redshift_lens,
            big_magnification_ones,
            self.magnification_map.convergence,
            self.magnification_map.shear,
            mean_microlens_mass_in_kg=self.magnification_map.mean_microlens_mass_in_kg,
            total_microlens_einstein_radii=1,
            name="identity",
        )

        wavelength_1 = 100
        wavelength_2 = 300

        micro_tf_1_id = (
            self.identity_magnification_array.calculate_microlensed_transfer_function(
                self.AccretionDisk,
                wavelength_1,
            )
        )

        micro_tf_2_id = (
            self.identity_magnification_array.calculate_microlensed_transfer_function(
                self.AccretionDisk,
                wavelength_2,
            )
        )

        tf_1 = self.AccretionDisk.construct_accretion_disk_transfer_function(
            wavelength_1
        )

        tf_2 = self.AccretionDisk.construct_accretion_disk_transfer_function(
            wavelength_2
        )

        assert len(micro_tf_1_id) == len(micro_tf_2_id)
        assert len(tf_1) == len(tf_2)

        tau_ax = np.linspace(0, len(tf_1) - 1, len(tf_1))
        tau_ax_ml = np.linspace(0, len(micro_tf_1_id) - 1, len(micro_tf_1_id))

        mean_micro_tf_1_id = np.sum(tau_ax_ml * micro_tf_1_id)
        mean_micro_tf_2_id = np.sum(tau_ax_ml * micro_tf_2_id)
        mean_tf_1 = np.sum(tau_ax * tf_1)
        mean_tf_2 = np.sum(tau_ax * tf_2)

        npt.assert_almost_equal(np.sum(micro_tf_1_id), 1)
        npt.assert_almost_equal(np.sum(micro_tf_2_id), 1)

        npt.assert_almost_equal(round(mean_micro_tf_1_id, 0), round(mean_tf_1, 0), 1)
        npt.assert_almost_equal(round(mean_micro_tf_2_id, 0), round(mean_tf_2, 0), 1)
        assert mean_micro_tf_2_id > mean_micro_tf_1_id



import pytest
import numpy as np
import numpy.testing as npt
from amoeba.Classes.blr_streamline import Streamline


def test_init():

    launch_radius = 500  # Rg
    launch_theta = 0  # degrees
    max_height = 1000  # Rg
    characteristic_distance = max_height // 5
    asymptotic_poloidal_velocity = 0.2
    poloidal_launch_velocity = 10**-5
    too_fast = 1.2
    too_close = 0
    too_wide = 90
    bad_radial_vector = np.linspace(0, 100, 100)
    bad_velocity_vector = np.linspace(0, 0.4, 20)

    with npt.assert_raises(AssertionError):
        Streamline(
            launch_radius,
            launch_theta,
            max_height,
            characteristic_distance,
            too_fast,
        )

    with npt.assert_raises(AssertionError):
        Streamline(
            too_close,
            launch_theta,
            max_height,
            characteristic_distance,
            asymptotic_poloidal_velocity,
        )

    with npt.assert_raises(AssertionError):
        Streamline(
            launch_radius,
            too_wide,
            max_height,
            characteristic_distance,
            asymptotic_poloidal_velocity,
        )

    with npt.assert_raises(AssertionError):
        Streamline(
            launch_radius,
            launch_theta,
            max_height,
            characteristic_distance,
            asymptotic_poloidal_velocity,
            poloidal_launch_velocity=too_fast,
        )

    with npt.assert_raises(AssertionError):
        Streamline(
            launch_radius,
            launch_theta,
            max_height,
            characteristic_distance,
            asymptotic_poloidal_velocity,
            poloidal_launch_velocity=too_fast,
        )

    with npt.assert_raises(AssertionError):
        Streamline(
            launch_radius,
            launch_theta,
            max_height,
            characteristic_distance,
            asymptotic_poloidal_velocity,
            velocity_vector=bad_velocity_vector,
            radial_vector=bad_radial_vector,
        )

    test_blr_streamline = Streamline(
        launch_radius,
        launch_theta,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        poloidal_launch_velocity=poloidal_launch_velocity,
    )

    manual_radii = np.linspace(20, 100, 101)
    manual_velocities = 0.1 + 0.05 * np.sin(np.linspace(0, 10, 101) / np.pi)

    strange_but_okay_streamline = Streamline(
        launch_radius,
        launch_theta,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        velocity_vector=manual_velocities,
        radial_vector=manual_radii,
    )

    with npt.assert_raises(AssertionError):
        Streamline(
            launch_radius,
            launch_theta,
            max_height,
            characteristic_distance,
            asymptotic_poloidal_velocity,
            velocity_vector=manual_velocities,
            radial_vector=bad_radial_vector,
        )

    strange_but_differently_overridden_streamline = Streamline(
        launch_radius,
        launch_theta,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        radial_vector=manual_radii,
    )

    automatic_radii_streamline = Streamline(
        launch_radius,
        launch_theta,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        velocity_vector=manual_velocities,
    )

    assert test_blr_streamline.launch_radius == launch_radius
    assert test_blr_streamline.launch_theta == launch_theta * np.pi / 180
    assert test_blr_streamline.launch_height > 0
    assert test_blr_streamline.poloidal_launch_velocity == poloidal_launch_velocity
    assert test_blr_streamline.height_step == 10  # default value
    assert test_blr_streamline.poloidal_launch_velocity == poloidal_launch_velocity
    assert (
        test_blr_streamline.radial_launch_velocity
        == poloidal_launch_velocity * np.sin(test_blr_streamline.launch_theta)
    )
    assert np.size(test_blr_streamline.height_values) == np.size(
        test_blr_streamline.poloidal_velocity
    )

    # velocities should be increasing!
    # radii should be constant!
    for jj in range(len(test_blr_streamline.poloidal_velocity) - 1):
        assert (
            test_blr_streamline.poloidal_velocity[jj + 1]
            > test_blr_streamline.poloidal_velocity[jj]
        )
        assert (
            test_blr_streamline.radii_values[jj + 1]
            == test_blr_streamline.radii_values[jj]
        )

    # choose a nice angle to relate max radius to max height
    launch_theta_angled = 45
    test_blr_streamline_angled = Streamline(
        launch_radius,
        launch_theta_angled,
        max_height,
        characteristic_distance,
        asymptotic_poloidal_velocity,
        poloidal_launch_velocity=poloidal_launch_velocity,
    )

    for jj in range(len(test_blr_streamline_angled.poloidal_velocity) - 1):
        assert (
            test_blr_streamline_angled.poloidal_velocity[jj + 1]
            > test_blr_streamline_angled.poloidal_velocity[jj]
        )
        assert (
            test_blr_streamline_angled.radii_values[jj + 1]
            > test_blr_streamline_angled.radii_values[jj]
        )
        print(test_blr_streamline_angled.poloidal_velocity)
        print(test_blr_streamline_angled.dpol_vel_dz_on_vel)
        assert test_blr_streamline_angled.dpol_vel_dz_on_vel[jj + 1] > 0

    assert (
        test_blr_streamline_angled.radii_values[-1] - test_blr_streamline.launch_radius
        == test_blr_streamline_angled.height_values[-1]
    )
